"""
SpatioTempSpectra (:mod:`fluidsim.solvers.ns2d.strat.output.spatio_temporal_spectra`)
=====================================================================================


Provides:

.. autoclass:: SpatioTempSpectra
   :members:
   :private-members:

"""

import sys
import time
import numpy as np
import h5py
import math
import matplotlib.pyplot as plt

from math import pi
from scipy import signal
from pathlib import Path

from fluiddyn.util import mpi
from fluidsim import load_state_phys_file
from fluidsim.base.output.base import SpecificOutput

# Notes
# -----
# Use scipy.signal.periodogram
# https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.signal.periodogram.html

# Lindborg, Brethouwer 2007 "Stratified turbulence forced in rotational and divergent modes"


class SpatioTempSpectra(SpecificOutput):
    """
    Compute, save, load and plot the spatio temporal spectra.

    """

    _tag = "spatio_temporal_spectra"
    _name_file = _tag + ".h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "spatio_temporal_spectra"

        params.output.periods_save._set_attrib(tag, False)
        params.output._set_child(
            tag,
            attribs={
                "HAS_TO_PLOT_SAVED": False,
                "time_start": 1,
                "time_decimate": 2,
                "spatial_decimate": 2,
                "size_max_file": 0.1,
                "kx_max": None,
                "kz_max": None,
            },
        )

    def __init__(self, output):
        params = output.sim.params
        pspatiotemp_spectra = params.output.spatio_temporal_spectra
        super().__init__(
            output,
            period_save=params.output.periods_save.spatio_temporal_spectra,
            has_to_plot_saved=pspatiotemp_spectra.HAS_TO_PLOT_SAVED,
        )

        # Parameters
        self.time_start = pspatiotemp_spectra.time_start
        self.time_decimate = pspatiotemp_spectra.time_decimate
        self.spatial_decimate = pspatiotemp_spectra.spatial_decimate
        self.size_max_file = pspatiotemp_spectra.size_max_file

        self.kx_max = pspatiotemp_spectra.kx_max
        self.kz_max = pspatiotemp_spectra.kz_max

        # By default: kxmax_dealiasing or kymax_dealiasing
        if self.kx_max is None:
            self.kx_max = self.sim.oper.kxmax_dealiasing

        if self.kz_max is None:
            self.kz_max = self.sim.oper.kymax_dealiasing

        self.has_to_save = bool(
            params.output.periods_save.spatio_temporal_spectra
        )

        # Check: In restart... time_start == time last state.
        path_dir = Path(self.sim.output.path_run)
        path_spatio_temp_files = path_dir / "spatio_temporal"

        if sorted(path_spatio_temp_files.glob("spatio_temp*")):
            time_last_file = float(
                sorted(path_dir.glob("state_phys*"))[-1].stem.split(
                    "state_phys_t"
                )[1]
            )

            if round(self.time_start, 3) != time_last_file:
                self.time_start = time_last_file

        # Compute arrays
        nK0, nK1 = self.sim.oper.shapeK_seq

        # The maximum value kx_max should be the dealiased value
        if self.kx_max > self.sim.oper.kxmax_dealiasing:
            self.kx_max = self.sim.oper.kxmax_dealiasing

        # The maximum value kz_max should be the dealiased value
        if self.kz_max > self.sim.oper.kymax_dealiasing:
            self.kz_max = self.sim.oper.kymax_dealiasing

        # Modified values to take into account kx_max and kz_max
        self.nK0 = np.argmin(abs(self.sim.oper.kxE - self.kx_max))
        self.nK1 = np.argmin(abs(self.sim.oper.kyE - self.kz_max))

        nK0_dec = len(list(range(0, self.nK0, self.spatial_decimate)))
        nK1_dec = len(list(range(0, self.nK1, self.spatial_decimate)))

        # Compute size in bytes of one array
        # self.size_max_file is given in Mbytes. 1 Mbyte == 1024 ** 2 bytes
        nb_bytes = np.empty([nK0_dec, nK1_dec], dtype=complex).nbytes
        self.nb_arr_in_file = int(self.size_max_file * (1024 ** 2) // nb_bytes)
        mpi.printby0("nb_arr_in_file", self.nb_arr_in_file)

        # Check: duration file <= duration simulation
        self.duration_file = (
            self.nb_arr_in_file
            * self.params.time_stepping.deltat0
            * self.time_decimate
        )
        if (
            self.duration_file > self.params.time_stepping.t_end
            and self.has_to_save
        ):
            raise ValueError(
                "The duration of the simulation is not enough to fill a file."
            )

        # Check: self.nb_arr_in_file should be > 0
        if self.nb_arr_in_file <= 0 and self.has_to_save:
            raise ValueError("The size of the file should be larger.")

        else:
            # Create array 4D (2 keys, times, n0, n1)

            self.spatio_temp_new = np.empty(
                [2, self.nb_arr_in_file, nK0_dec, nK1_dec], dtype=complex
            )

        #  Convert time_start to it_start.
        if not self.sim.time_stepping.it:
            self.it_start = int(
                self.time_start / self.params.time_stepping.deltat0
            )
        else:
            # If simulation starts from a specific time...
            self.it_start = self.sim.time_stepping.it + int(
                (self.time_start - round(self.sim.time_stepping.t, 3))
                / self.params.time_stepping.deltat0
            )

            if self.it_start < self.sim.time_stepping.it:
                self.it_start = self.sim.time_stepping.it

        # Create empty array with times
        self.times_arr = np.empty([self.nb_arr_in_file])
        self.its_arr = np.empty([self.nb_arr_in_file])

        if params.time_stepping.USE_CFL and self.has_to_save:
            raise ValueError(
                "To compute the spatio temporal: \n"
                "USE_CFL = FALSE and periods_save.spatio_temporal_spectra > 0"
            )

        # Create directory to save files
        dir_name = "spatio_temporal"
        self.path_dir = Path(self.sim.output.path_run) / dir_name

        if self.has_to_save and mpi.rank == 0:
            self.path_dir.mkdir(exist_ok=True)

        # Start loop in _online_save
        self.it_last_run = self.it_start
        self.nb_times_in_spatio_temp = 0

    def _init_files(self, arrays_1st_time=None):
        # we can not do anything when this function is called.
        pass

    def _write_to_file(self, spatio_temp_arr, times_arr, its_arr):
        """Writes a file with the spatio temporal data"""
        if mpi.rank == 0:
            # Name file
            it_end = self.sim.time_stepping.it
            name_file = f"spatio_temp_it{it_end}.h5"
            path_file = self.path_dir / name_file

            # Dictionary arrays
            dict_arr = {
                "its_arr": its_arr,
                "times_arr": times_arr,
                "spatio_temp": spatio_temp_arr,
            }

            # Write dictionary to file
            with h5py.File(path_file, "w") as file:
                for k, v in list(dict_arr.items()):
                    file.create_dataset(k, data=v)

    def _online_save(self):
        """Computes and saves the values at one time."""
        oper = self.sim.oper
        if not self.has_to_save:
            return

        itsim = self.sim.time_stepping.it
        if itsim - self.it_last_run >= self.time_decimate:
            self.it_last_run = itsim

            # Save the field to self.spatio_temp
            field_ap = self.sim.state.compute("ap_fft")
            field_am = self.sim.state.compute("am_fft")

            if mpi.nb_proc > 1:
                # Create big array
                if mpi.rank == 0:
                    field_ap_seq = oper.create_arrayK(shape="seq")
                    field_am_seq = oper.create_arrayK(shape="seq")
                else:
                    field_ap_seq = None
                    field_am_seq = None

                # Define size of each array
                sendcounts = np.array(mpi.comm.gather(field_ap.size, root=0))

                # Send array each process to root (process 0)
                mpi.comm.Gatherv(
                    sendbuf=field_ap, recvbuf=(field_ap_seq, sendcounts), root=0
                )
                mpi.comm.Gatherv(
                    sendbuf=field_am, recvbuf=(field_am_seq, sendcounts), root=0
                )

            else:
                field_ap_seq = field_ap
                field_am_seq = field_am

            # Cut off kx_max and kz_max
            if mpi.rank == 0:
                field_ap_seq = field_ap_seq[0 : self.nK0, 0 : self.nK1]
                field_am_seq = field_am_seq[0 : self.nK0, 0 : self.nK1]

            # Decimate with process 0
            if mpi.rank == 0:
                field_ap_decimate = field_ap_seq[
                    :: self.spatial_decimate, :: self.spatial_decimate
                ]
                field_am_decimate = field_am_seq[
                    :: self.spatial_decimate, :: self.spatial_decimate
                ]
                self.spatio_temp_new[
                    0, self.nb_times_in_spatio_temp, :, :
                ] = field_ap_decimate
                self.spatio_temp_new[
                    1, self.nb_times_in_spatio_temp, :, :
                ] = field_am_decimate

            # Save the time to self.times_arr
            if not self.sim.time_stepping.it:
                self.times_arr[self.nb_times_in_spatio_temp] = (
                    itsim * self.sim.params.time_stepping.deltat0
                )
            else:
                self.times_arr[
                    self.nb_times_in_spatio_temp
                ] = self.sim.time_stepping.t

            self.its_arr[self.nb_times_in_spatio_temp] = self.sim.time_stepping.it

            # Check if self.spatio_temp is filled. If yes, writes to a file.
            if (
                self.nb_times_in_spatio_temp == self.nb_arr_in_file - 1
                or self.sim.time_stepping.is_simul_completed() == True
            ):
                if mpi.rank == 0:
                    print("Saving spatio_temporal data...")

                # If the file is not completed AND the simulation is finished, it saves the file too.
                # Check if the last time is good.
                if self.sim.time_stepping.is_simul_completed() and mpi.rank == 0:
                    self.spatio_temp_new = self.spatio_temp_new[
                        :, : self.nb_times_in_spatio_temp + 1, :, :
                    ]
                    self.times_arr = self.times_arr[
                        : self.nb_times_in_spatio_temp + 1
                    ]
                    self.its_arr = self.its_arr[
                        : self.nb_times_in_spatio_temp + 1
                    ]

                self._write_to_file(
                    self.spatio_temp_new, self.times_arr, self.its_arr
                )

                self.nb_times_in_spatio_temp = 0
            else:
                self.nb_times_in_spatio_temp += 1

    def compute_frequency_spectra(
        self, tmin=None, tmax=None, time_windows=None, overlap=0.5
    ):
        """Computes the temporal frequency spectrum.

        Keyword Arguments:
            tmin {float} -- Minimum time to perform temporal FT. (default: {None})
            tmax {float} -- Maximum time to perform temporal FT. (default: {None})
            time_windows {float} -- Time windows to compute temporal FT. (default: {None})
            overlap {float} -- Overlap between windows. From 0 to 1. (default: {0.5})
        """

        # Get concatenated array of times and spatio temporal
        times, spatio_temp = self._get_concatenate_data(tmin, tmax)

        # Compute windows_size
        if time_windows is None:
            nb_periods_N = 10
            windows_size = int(
                2
                * pi
                / self.sim.params.N
                * nb_periods_N
                * (1 / self.sim.params.time_stepping.deltat0 / self.time_decimate)
            )
        else:
            windows_size = int(
                time_windows
                * (1 / self.sim.params.time_stepping.deltat0 / self.time_decimate)
            )

        # # Compute time_windows
        # if not nb_periods_N:
        # nb_periods_N = 10

        # windows_size = int(
        # 2
        # * pi
        # / self.sim.params.N
        # * nb_periods_N
        # * (1 / self.sim.params.time_stepping.deltat0 / self.time_decimate)
        # )

        if windows_size > len(times):
            windows_size = len(times)

        # Check argument overlap
        if overlap < 0 or overlap > 1:
            raise ValueError("Overlap should be between 0 and 1.")

        # Compute number of indexes to skip
        skip = int((1 - overlap) * windows_size)

        # Slice step cannot be zero
        if not skip or skip == 1:
            skip = 2

        # Array of times to start the FT
        times_start = times[:: skip - 1]
        times_start = times_start[:-1]

        # Create array to allocate all FT's
        spatio_temp_all = np.empty(
            [
                len(times_start),
                2,
                windows_size,
                spatio_temp.shape[2],
                spatio_temp.shape[3],
            ],
            dtype=complex,
        )

        # Compute sampling frequency in Hz
        freq_sampling = 1.0 / (
            self.time_decimate * self.params.time_stepping.deltat0
        )

        # Compute temporal FT
        for index, time_start in enumerate(times_start):

            # Compute index to start and to end FT
            it0 = np.argmin(abs(times - time_start))
            it1 = it0 + windows_size

            # Compute spectrum
            omegas, temp_spectrum = signal.periodogram(
                spatio_temp[:, it0:it1, :, :],
                fs=freq_sampling,
                window="hann",
                nfft=windows_size,
                detrend="constant",
                return_onesided=False,
                scaling="spectrum",
                axis=1,
            )

            # Allocate FT
            spatio_temp_all[index, :, :, :, :] = temp_spectrum

        # Build up a dictionary with the rsults
        dict_arr = {
            "omegas": omegas,
            "spectrum": np.mean(spatio_temp_all, axis=0),
            "windows_size": windows_size,
        }

        # Create name path file
        str_time = time.strftime("%Y_%m_%d-%H_%M_%S")
        name_file = f"spectrum_{str_time}.h5"

        # Create directory if does not exist
        path_dir = Path(self.path_dir) / "Spectrum"
        path_dir.mkdir(exist_ok=True)

        path_file = path_dir / name_file

        # Save file into directory spatio_temporal/Spectrum
        with h5py.File(path_file, "w") as file:
            for k, v in list(dict_arr.items()):
                file.create_dataset(k, data=v)

    def _get_concatenate_data(self, tmin, tmax):
        """Gives concatenated time and spatio_temp arrays/

        Arguments:
            tmin {float} -- Minimum time to perform temporal FT.
            tmax {float} -- Maximum time to perform temporal FT.

        Raises:
            ValueError -- itmin == itmax

        Returns:
            times {array} -- Concatenated time array
            spatio_temp {array} -- concatenated spatio temporal array
        """

        # Define list files
        list_files = sorted(self.path_dir.glob("spatio_temp_it*"))

        # Sort files by number of iteration.
        list_its = []

        for path_file in list_files:
            list_its.append(int(path_file.name.split("_it")[1].split(".h5")[0]))

        list_its = sorted(list_its)
        print(list_its)
        list_files_new = []
        for it in list_its:
            list_files_new.append(
                self.path_dir / ("spatio_temp_it" + str(it) + ".h5")
            )

        list_files = list_files_new

        # List of tuples (ifile, time)
        ifile_time = []

        for index, path_file in enumerate(list_files):
            with h5py.File(path_file, "r") as file:
                times = file["times_arr"][...]
            for time_value in times:
                ifile_time.append((index, time_value))

        ifile_min = None
        ifile_max = None

        for ii, value in enumerate(ifile_time):
            if ifile_min is None and tmin == round(value[1], 0):
                ifile_min = value[0]
            if ifile_max is None and tmax == round(value[1], 0):
                ifile_max = value[0]

        if ifile_min is None:
            ifile_min = 0

        if ifile_max is None:
            ifile_max = len(list_files)

        # Define concatenate arrays
        spatio_temp_conc = None
        times_conc = None

        # Load all data
        for index, path_file in enumerate(list_files[ifile_min:ifile_max]):
            with h5py.File(path_file, "r") as file:
                spatio_temp = file["spatio_temp"][...]
                times = file["times_arr"][...]

            # Print concatenating info..
            print(
                f"Concatenating file = {index + 1}/{len(list_files[ifile_min:ifile_max])}..",
                end="\r",
            )

            # Concatenate arrays
            if isinstance(spatio_temp_conc, np.ndarray):
                spatio_temp_conc = np.concatenate(
                    (spatio_temp_conc, spatio_temp), axis=1
                )
            if isinstance(times_conc, np.ndarray):
                times_conc = np.concatenate((times_conc, times), axis=0)
            else:
                spatio_temp_conc = spatio_temp
                times_conc = times

            sys.stdout.flush()
            time.sleep(0.2)

        # Default tmin and tmax
        if not tmin:
            tmin = np.min(times_conc)
        if not tmax:
            tmax = np.max(times_conc)

        # Find indexes itmin and itmax
        itmin = np.argmin(abs(times_conc - tmin))
        itmax = np.argmin(abs(times_conc - tmax)) + 1

        # Check..
        if itmin == itmax:
            raise ValueError(
                "itmin == itmax. \n"
                + f"Choose between tmin = {np.min(times_conc)} - "
                + f"tmax = {np.max(times_conc)}"
            )

        # Print info..
        to_print = f"tmin = {np.min(times_conc)} ; tmax = {np.max(times_conc)}"
        print(to_print)

        return times_conc[itmin:itmax], spatio_temp_conc[:, itmin:itmax, :, :]

    def _add_inset_plot(self, fig, kxmax_inset=32, kzmax_inset=32):
        """
        Adds an inset plot to the object figure.
        It is the spectral space with the forcing region.
        """
        # Checks if self.sim.forcing.forcing_maker object exists.
        if self.sim.forcing.forcing_maker is None:
            ax_inset = None
        else:
            # Define position of the inset plot
            # There are percentages of the original axes.
            left, bottom, width, height = [0.56, 0.68, 0.22, 0.22]
            ax_inset = fig.add_axes([left, bottom, width, height])

            # Add labels
            ax_inset.set_xlabel("$k_x$", labelpad=0.05)
            ax_inset.set_ylabel("$k_z$", labelpad=0.05)

            # Set ticks each axis
            xticks = np.arange(0, kxmax_inset, self.sim.oper.deltakx)
            zticks = np.arange(0, kzmax_inset, self.sim.oper.deltaky)
            ax_inset.set_xticks(xticks[:: int(self.sim.oper.deltaky)])
            ax_inset.set_yticks(zticks)

            # Set axis limit & grid
            ax_inset.set_xlim(0, kxmax_inset)
            ax_inset.set_ylim(0, kzmax_inset)
            ax_inset.grid(linestyle="--", alpha=0.4)

            indices_forcing = np.argwhere(
                self.sim.forcing.forcing_maker.COND_NO_F == False
            )

            for i, index in enumerate(indices_forcing):
                ax_inset.plot(
                    self.sim.oper.KX[0, index[1]],
                    self.sim.oper.KY[index[0], 0],
                    color="red",
                    marker="o",
                    markersize=2,
                    label="Forced mode" if i == 0 else "",
                )

        return ax_inset

    def plot_kx_omega_cross_section(
        self,
        path_file=None,
        field="ap_fft",
        ikz_plot=None,
        kxmax_plot=None,
        func_plot="pcolormesh",
        INSET_PLOT=True,
    ):
        pspatio = self.params.output.spatio_temporal_spectra
        # Define path dir as posix path
        path_dir = self.path_dir

        # If path_file does not exist.
        if not path_file:
            path_file = sorted((path_dir / "Spectrum").glob("spectrum*"))[-1]

        print(f"File = {path_file}")
        # Load data
        with h5py.File(path_file, "r") as file:
            temp_spectrum_mean = file["spectrum"][...]
            omegas = file["omegas"][...]

        # Load simulation object
        sim = load_state_phys_file(path_dir.parent, merge_missing_params=True)

        # Cut_off indexes
        # ikx_top = np.argmin(abs(sim.oper.kx - sim.oper.kxmax_dealiasing))
        # ikz_top = np.argmin(abs(sim.oper.ky - sim.oper.kymax_dealiasing))

        kx_max = pspatio.kx_max
        kz_max = pspatio.kx_max

        if kx_max is None:
            kx_max = sim.oper.kxmax_dealiasing

        if kz_max is None:
            kz_max = sim.oper.kymax_dealiasing

        print("kx_max", kx_max)

        ikx_top = np.argmin(abs(sim.oper.kx - kx_max))
        ikz_top = np.argmin(abs(sim.oper.ky - kz_max))

        # Cut_off kx_max and ky_max
        kx_cut_off = sim.oper.kx[0:ikx_top]
        kz_cut_off = sim.oper.ky[0:ikz_top]

        # Compute kx_decimate and ky_decimate
        kx_decimate = kx_cut_off[
            :: sim.params.output.spatio_temporal_spectra.spatial_decimate
        ]
        kz_decimate = kz_cut_off[
            :: sim.params.output.spatio_temporal_spectra.spatial_decimate
        ]

        # Omegas
        omegas *= 2 * np.pi
        omegas *= 1 / (sim.params.N)

        #### PLOT OMEGA - KX

        # Define maximum and minimum values to plot.
        kxmin_plot = 0
        if kxmax_plot is None:
            kxmax_plot = sim.oper.kx[ikx_top]
        else:
            ikxmax_plot = np.argmin(abs(sim.oper.kx - kxmax_plot))
            kxmax_plot = sim.oper.kx[ikxmax_plot]

        ikxmin_plot = np.argmin(abs(kx_decimate - kxmin_plot))
        ikxmax_plot = np.argmin(abs(kx_decimate - kxmax_plot)) + 1

        omega_max_plot = 0
        omega_min_plot = -3

        omegas = np.append(omegas, 0)
        imin_omegas_plot = np.argmin(abs(omegas - omega_min_plot))

        kxs_grid, omegas_grid = np.meshgrid(
            kx_decimate[ikxmin_plot:ikxmax_plot], omegas[imin_omegas_plot:]
        )

        if ikz_plot is None:
            ikz_plot = 1

        # Parameters figure
        fig, ax = plt.subplots()
        ax.set_xlim(left=kxmin_plot, right=kxmax_plot)
        ax.set_ylim(top=omega_max_plot, bottom=omega_min_plot)
        ax.set_xlabel(r"$k_x$", fontsize=16)
        ax.set_ylabel(r"$\omega / N$", fontsize=16)
        ax.tick_params(axis="x", labelsize=16)
        ax.tick_params(axis="y", labelsize=16)
        ax.set_title(fr"$log_{{10}} E(k_x, {kz_decimate[ikz_plot]}, \omega)$")

        # Compute index array corresponding to field.
        if field == "ap_fft":
            i_field = 0
        elif field == "am_fft":
            i_field = 1
        else:
            raise ValueError(f"field = {field} not known.")

        # Compute data
        temp_spectrum_mean = temp_spectrum_mean[i_field]

        # todo: fix the warnings that we get while testing
        # data = np.log10(
        # temp_spectrum_mean[imin_omegas_plot:, ikz_plot, ikxmin_plot:ikxmax_plot]
        # )

        data = np.log10(
            temp_spectrum_mean[
                imin_omegas_plot:, ikxmin_plot:ikxmax_plot, ikz_plot
            ]
        )

        new = np.empty(
            [
                temp_spectrum_mean.shape[0] + 1,
                temp_spectrum_mean.shape[1],
                temp_spectrum_mean.shape[2],
            ]
        )
        new[:-1] = temp_spectrum_mean
        new[-1] = temp_spectrum_mean[-1]
        data = np.log10(new[imin_omegas_plot:, ikxmin_plot:ikxmax_plot, ikz_plot])
        # data = np.log10(new[imin_omegas_plot:, ikz_plot, ikxmin_plot:ikxmax_plot])

        # Maximum and minimum values
        vmin = np.min(data[abs(data) != np.inf])
        vmax = np.max(data)

        # Plot with contourf
        import matplotlib.cm as cm

        if func_plot == "contourf":
            cf = ax.contourf(
                kxs_grid, omegas_grid, data, cmap=cm.viridis, vmin=vmin, vmax=vmax
            )
        elif func_plot == "pcolormesh":
            # To fit in pcolormesh
            kxs_grid = kxs_grid - (sim.oper.deltakx / 2)
            omegas_grid = omegas_grid - (omegas[1] / 2)

            cf = ax.pcolormesh(
                kxs_grid,
                omegas_grid,
                data,
                shading="nearest",
                cmap=cm.viridis,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            print(f"Function plot not known.")

        # Plot colorbar
        m = plt.cm.ScalarMappable(cmap=cm.viridis)
        m.set_array(data)
        m.set_clim(vmin, vmax)
        plt.colorbar(m)

        # Plot dispersion relation
        ax.plot(
            kxs_grid[0],
            -(
                kxs_grid[0]
                / np.sqrt(kxs_grid[0] ** 2 + kz_decimate[ikz_plot] ** 2)
            ),
            color="k",
        )

        # Check if angle string or float instance.
        if (
            isinstance(sim.params.forcing.tcrandom_anisotropic.angle, str)
            and "°" in sim.params.forcing.tcrandom_anisotropic.angle
        ):
            angle = math.radians(
                float(sim.params.forcing.tcrandom_anisotropic.angle.split("°")[0])
            )
        else:
            angle = sim.params.forcing.tcrandom_anisotropic.angle

        # Plot forcing region
        ax.axvline(
            sim.oper.deltaky * sim.params.forcing.nkmin_forcing * np.sin(angle),
            color="red",
        )

        ax.axvline(
            sim.oper.deltaky * sim.params.forcing.nkmax_forcing * np.sin(angle),
            color="red",
        )

        # Plot text shear modes
        ax.text(
            sim.oper.deltakx // 2,
            -1,
            r"$E(k_x=0, k_z) = 0$",
            rotation=90,
            fontsize=14,
            color="white",
        )

        # Inset plot
        if INSET_PLOT:
            ax_inset = self._add_inset_plot(fig)
            if ax_inset is not None:
                ax_inset.axhline(y=self.sim.oper.ky[ikz_plot], color="k")

        # Tight layut of the figure
        fig.tight_layout()

    def plot_kz_omega_cross_section(
        self,
        path_file=None,
        field="ap_fft",
        ikx_plot=None,
        func_plot="pcolormesh",
        INSET_PLOT=True,
    ):
        # if ax_inset is not None:
        #  D    efine path dir as po   six path
        path_dir = self.path_dir
        # If path_file does not exist.
        if not path_file:
            path_file = sorted((path_dir / "Spectrum").glob("spectrum*"))[-1]

        # Print path_file
        print(f"Path = {path_file}")

        # Load data
        with h5py.File(path_file, "r") as file:
            temp_spectrum_mean = file["spectrum"][...]
            omegas = file["omegas"][...]

        # Load simulation object
        sim = load_state_phys_file(path_dir.parent, merge_missing_params=True)

        # Compute kx_decimate and ky_decimate
        kx_decimate = sim.oper.kx[
            :: sim.params.output.spatio_temporal_spectra.spatial_decimate
        ]
        kz_decimate = sim.oper.ky[
            :: sim.params.output.spatio_temporal_spectra.spatial_decimate
        ]

        # Omegas
        omegas *= 2 * np.pi
        omegas *= 1 / (sim.params.N)

        #### PLOT OMEGA - KZ
        kzmin_plot = 0
        kzmax_plot = 80

        if kzmax_plot > sim.oper.kymax_dealiasing:
            kzmax_plot = sim.oper.kymax_dealiasing

        ikzmin_plot = np.argmin(abs(kz_decimate - kzmin_plot))
        ikzmax_plot = np.argmin(abs(kz_decimate - kzmax_plot))

        omega_max_plot = 0
        omega_min_plot = -3

        omegas = np.append(omegas, 0)
        imin_omegas_plot = np.argmin(abs(omegas - omega_min_plot)) - 1

        kzs_grid, omegas_grid = np.meshgrid(
            kz_decimate[ikzmin_plot:ikzmax_plot], omegas[imin_omegas_plot:]
        )

        if not ikx_plot:
            ikx_plot = 1

        # Parameters figure
        fig, ax = plt.subplots()
        ax.set_xlim(left=kzmin_plot, right=kzmax_plot - sim.oper.deltaky / 2)
        ax.set_ylim(top=omega_max_plot, bottom=omega_min_plot)
        ax.set_xlabel(r"$k_z$", fontsize=16)
        ax.set_ylabel(r"$\omega / N$", fontsize=16)
        ax.tick_params(axis="x", labelsize=16)
        ax.tick_params(axis="y", labelsize=16)
        ax.set_title(fr"$log_{{10}} E({kx_decimate[ikx_plot]}, k_z, \omega)$")

        # Compute index array corresponding to field.
        if field == "ap_fft":
            i_field = 0
        elif field == "am_fft":
            i_field = 1
        else:
            raise ValueError(f"field = {field} not known.")

        # Compute data
        temp_spectrum_mean = temp_spectrum_mean[i_field]

        # todo: fix the warnings that we get while testing
        data = np.log10(
            temp_spectrum_mean[
                imin_omegas_plot:, ikx_plot, ikzmin_plot:ikzmax_plot
            ]
        )
        new = np.empty(
            [
                temp_spectrum_mean.shape[0] + 1,
                temp_spectrum_mean.shape[1],
                temp_spectrum_mean.shape[2],
            ]
        )
        new[:-1] = temp_spectrum_mean
        new[-1] = temp_spectrum_mean[-1]
        data = np.log10(new[imin_omegas_plot:, ikx_plot, ikzmin_plot:ikzmax_plot])

        # Maximum and minimum values
        vmin = np.min(data[abs(data) != np.inf])
        vmin = -3
        vmax = np.max(data)
        vmax = 0

        # Plot with contourf
        import matplotlib.cm as cm

        if func_plot == "contourf":
            cf = ax.contourf(
                kzs_grid, omegas_grid, data, cmap=cm.viridis, vmin=vmin, vmax=vmax
            )
        elif func_plot == "pcolormesh":
            kzs_grid = kzs_grid - (sim.oper.deltaky / 2)
            omegas_grid = omegas_grid - (omegas[1] / 2)
            cf = ax.pcolormesh(
                kzs_grid,
                omegas_grid,
                data,
                shading="nearest",
                cmap=cm.viridis,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            print(f"Function plot not known.")

        # Plot colorbar
        m = plt.cm.ScalarMappable(cmap=cm.viridis)
        m.set_array(data)
        m.set_clim(vmin, vmax)
        plt.colorbar(m)

        # Plot dispersion relation
        ax.plot(
            kzs_grid[0][ikzmin_plot:ikzmax_plot],
            -(
                kx_decimate[ikx_plot]
                / np.sqrt(
                    kx_decimate[ikx_plot] ** 2
                    + kzs_grid[0][ikzmin_plot:ikzmax_plot] ** 2
                )
            ),
            color="k",
        )

        # Check if angle string or float instance.
        if (
            isinstance(sim.params.forcing.tcrandom_anisotropic.angle, str)
            and "°" in sim.params.forcing.tcrandom_anisotropic.angle
        ):
            angle = math.radians(
                float(sim.params.forcing.tcrandom_anisotropic.angle.split("°")[0])
            )
        else:
            angle = sim.params.forcing.tcrandom_anisotropic.angle

        # Plot forcing region
        ax.axvline(
            sim.oper.deltaky * sim.params.forcing.nkmin_forcing * np.sin(angle),
            color="red",
        )

        ax.axvline(
            sim.oper.deltaky * sim.params.forcing.nkmax_forcing * np.sin(angle),
            color="red",
        )

        # Inset plot
        if INSET_PLOT:
            ax_inset = self._add_inset_plot(fig)
            if ax_inset is not None:
                ax_inset.axvline(x=self.sim.oper.kx[ikx_plot], color="k")

        # Tight layout of the figure
        fig.tight_layout()

    def print_info_frequency_spectra(self):
        """Print information frequency spectra. """

        print("""*** Info frequency spectra ***""")
        print("dt = ", self.params.time_stepping.deltat0)
        print("Decimation time = ", self.time_decimate)
        print("Number arrays in file = ", self.nb_arr_in_file)
        print("Duration file = ", self.duration_file)
        print(
            "Total number files simulation = ",
            int(
                (self.params.time_stepping.t_end - self.time_start)
                / self.duration_file
            ),
        )

    def plot_frequency_spectra_individual_mode(
        self, path_file=None, mode=None, INSET_PLOT=True
    ):
        """Plots the frequency spectra of an individual Fourier mode.

        Keyword Arguments:
            path_file {string} -- Path of the file.
            mode {tuple} -- Mode (kx, kz) to plot.

        """

        # Define path_dir as posix
        path_dir = Path(self.path_dir)

        # If path_file does not exists..
        if not path_file:
            path_file = sorted((path_dir / "Spectrum").glob("spectrum*"))[-1]

        # If mode does is not given..
        if not mode:
            kx = kz = 16
            mode = (kx, kz)

        # Load frequency spectra
        print(f"path_file = {path_file}")
        with h5py.File(path_file, "r") as file:
            spectrum = file["spectrum"][...]
            omegas = file["omegas"][...]

        # Define index with spatial decimation
        idx_mode = np.argmin(
            abs(self.sim.oper.kx[:: self.spatial_decimate] - mode[0])
        )
        idz_mode = np.argmin(
            abs(self.sim.oper.ky[:: self.spatial_decimate] - mode[1])
        )
        print(
            "kx_plot = {:.3f} ; kz_plot = {:.3f}".format(
                self.sim.oper.kx[:: self.spatial_decimate][idx_mode],
                self.sim.oper.ky[:: self.spatial_decimate][idz_mode],
            )
        )
        print(f"ikx_mode = {idx_mode} ; idz_mode = {idz_mode}")

        # Compute omega dispersion relation mode
        kx_mode = self.sim.oper.kx[:: self.spatial_decimate][idx_mode]
        kz_mode = self.sim.oper.ky[:: self.spatial_decimate][idz_mode]

        # Linear frequency. Used for compensation of the plots..
        N = self.sim.params.N
        f_l = N / (2 * np.pi)
        iw = (N / (2 * np.pi)) * kx_mode / (np.sqrt(kx_mode ** 2 + kz_mode ** 2))

        # Plot omega +
        fig1, ax1 = plt.subplots()
        ax1.set_xlabel(r"$\omega / N$")
        ax1.set_ylabel(r"$F(\omega)$")
        ax1.set_ylim(bottom=1e-7)
        ax1.set_title(
            r"$\omega_+ ; (k_x, k_z) = ({:.2f}, {:.2f})$".format(
                self.sim.oper.kx[:: self.spatial_decimate][idx_mode],
                self.sim.oper.ky[:: self.spatial_decimate][idz_mode],
            )
        )

        ax1.loglog(
            omegas[0 : len(omegas) // 2] / f_l,
            spectrum[0, 0 : len(omegas) // 2, idx_mode, idz_mode],
            "k",
        )
        ax1.axvline(x=f_l / f_l, color="k", linestyle="--")
        ax1.axvline(x=iw / f_l, color="r", linestyle="--")

        # Plot omega -
        fig2, ax2 = plt.subplots()
        ax2.set_xlabel(r"$\omega / N$")
        ax2.set_ylabel(r"$F(\omega)$")
        ax2.set_ylim(bottom=1e-7)
        ax2.set_title(
            r"$\omega_- ; (k_x, k_z) = ({:.2f}, {:.2f})$".format(
                self.sim.oper.kx[:: self.spatial_decimate][idx_mode],
                self.sim.oper.ky[:: self.spatial_decimate][idz_mode],
            )
        )

        ax2.loglog(
            -1 * omegas[len(omegas) // 2 + 1 :] / f_l,
            spectrum[0, len(omegas) // 2 + 1 :, idx_mode, idz_mode],
            "k",
        )
        ax2.axvline(x=f_l / f_l, color="k", linestyle="--")
        ax2.axvline(x=iw / f_l, color="r", linestyle="--")

        # Inset plot
        if INSET_PLOT:
            ax_inset = self._add_inset_plot(fig2)
            if ax_inset is not None:
                ax_inset.plot([kx_mode], [kz_mode], color="b", marker="o")

    def _init_online_plot(self):
        if mpi.rank == 0:
            fig, axe = self.output.figure_axe(numfig=1_000_000)
            self.axe = axe
            axe.set_xlabel("$k_h$")
            axe.set_ylabel("$E(k_h)$")
            axe.set_title("spectra\n" + self.output.summary_simul)



import unittest

import numpy as np

import fluiddyn.util.mpi as mpi


from fluidsim.util.testing import TestSimul, skip_if_no_fluidfft, classproperty


@skip_if_no_fluidfft
class TestSimulBase(TestSimul):
    @classproperty
    def Simul(cls):
        from fluidsim.solvers.plate2d.solver import Simul

        return Simul

    @classmethod
    def init_params(cls, Lh=2, nh=32):
        cls.params = params = cls.Simul.create_default_params()
        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"
        cls.nh = nh
        params.oper.nx = nh
        params.oper.ny = nh
        cls.Lh = Lh
        params.oper.Lx = Lh
        params.oper.Ly = Lh

        params.oper.coef_dealiasing = 2.0 / 3
        params.nu_8 = 2.0

        params.time_stepping.USE_CFL = False
        params.time_stepping.deltat0 = 0.005
        params.time_stepping.t_end = 0.5
        params.init_fields.type = "noise"
        params.output.HAS_TO_SAVE = False
        params.forcing.enable = False

        params.output.ONLINE_PLOT_OK = False
        return params


class TestSolverPlate2DTendency(TestSimulBase):
    def test_tendency(self):
        ratio = self.sim.test_tendencies_nonlin()
        self.assertGreater(2e-15, ratio)


@unittest.skipIf(mpi.nb_proc > 1, "plot function works sequentially only")
class TestSolverPlate2DInit(TestSimulBase):
    @classmethod
    def init_params(cls):
        params = super().init_params(4.0, nh=8)
        params.init_fields.type = "harmonic"
        params.init_fields.harmonic.i0 = 2
        params.init_fields.harmonic.i1 = 4

    def test_init(self):
        pass


class TestSolverPlate2DOutput(TestSimulBase):
    @classmethod
    def init_params(cls):
        params = super().init_params(6.0, nh=12)
        delta_x = cls.Lh / cls.nh

        kmax = np.sqrt(2) * np.pi / delta_x
        deltat = 2 * np.pi / kmax ** 2 / 2

        params.time_stepping.deltat0 = deltat
        params.time_stepping.it_end = 10
        params.time_stepping.USE_T_END = False

        params.init_fields.type = "noise"
        params.init_fields.noise.velo_max = 1e-6

        params.forcing.enable = True
        params.forcing.type = "tcrandom"
        params.forcing.forcing_rate = 1e4
        params.forcing.nkmax_forcing = 5
        params.forcing.nkmin_forcing = 2
        params.forcing.tcrandom.time_correlation = 100 * deltat

        params.nu_8 = (
            2e1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
        )
        params.output.HAS_TO_SAVE = True
        params.output.periods_print.print_stdout = 2 * deltat

        params.output.periods_save.phys_fields = 6 * deltat
        params.output.periods_save.spectra = 6 * deltat
        params.output.periods_save.spatial_means = 4 * deltat
        params.output.periods_save.correl_freq = 2 * deltat

        params.output.ONLINE_PLOT_OK = True
        params.output.period_refresh_plots = 6 * deltat

        params.output.spatial_means.HAS_TO_PLOT_SAVED = True
        params.output.spectra.HAS_TO_PLOT_SAVED = True

        params.output.correl_freq.HAS_TO_PLOT_SAVED = True
        params.output.correl_freq.it_start = 0
        params.output.correl_freq.nb_times_compute = 5
        params.output.correl_freq.coef_decimate = 1
        params.output.correl_freq.iomegas1 = [1, 2]

    def test_output(self):

        sim = self.sim

        sim.time_stepping.start()

        sim.output.correl_freq.compute_corr4_norm()

        for key in sim.info_solver.classes.State.keys_computable:
            sim.state.compute(key)

        sim.state.compute(key)

        with self.assertRaises(ValueError):
            sim.state.compute("abcdef")

        sim.state.compute("abcdef", RAISE_ERROR=False)

        if mpi.nb_proc > 1:
            return

        sim.output.correl_freq.plot_norm_pick_corr4()
        sim.output.correl_freq.plot_convergence()
        sim.output.correl_freq.plot_corr2()
        sim.output.correl_freq.plot_corr2_1d()
        sim.output.correl_freq.plot_corr4()
        sim.output.correl_freq.plot_convergence2()

        sim.output.spatial_means.plot(with_dtE=True)

        sim.output.spectra.plot1d()
        sim.output.spectra.plot2d()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# coding=utf8
#
# run modif_resolution.py
# python modif_resolution.py

from solveq2d import solveq2d

name_dir = (
    "~/Storage/Results_SW1lw/Pure_standing_waves_3840x3840/"
    + "SE2D_SW1lwaves_forcingw_L=50.x50._3840x3840_c=40_f=0_2013-07-23_15-47-25"
)
t_approx = 10.0e10
coef_modif_resol = 2

solveq2d.modif_resolution_from_dir(
    name_dir, t_approx=t_approx, coef_modif_resol=coef_modif_resol, PLOT=False
)



"""Energy budget (:mod:`fluidsim.solvers.ns2d.strat.output.spect_energy_budget`)
================================================================================

.. autoclass:: SpectralEnergyBudgetNS2DStrat
   :members:
   :private-members:

"""

import numpy as np
import h5py

from fluiddyn.util import mpi

from fluidsim.base.output.spect_energy_budget import (
    SpectralEnergyBudgetBase,
    cumsum_inv,
)


class SpectralEnergyBudgetNS2DStrat(SpectralEnergyBudgetBase):
    """Save and plot energy budget in spectral space."""

    def compute(self):
        """compute the spectral energy budget at one time."""
        oper = self.sim.oper

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")

        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        b_fft = self.sim.state.state_spect.get_var("b_fft")
        ux_fft, uy_fft = oper.vecfft_from_rotfft(rot_fft)

        px_b_fft, py_b_fft = oper.gradfft_from_fft(b_fft)
        px_b = oper.ifft2(px_b_fft)
        py_b = oper.ifft2(py_b_fft)

        Fb = -ux * px_b - uy * py_b
        Fb_fft = oper.fft2(Fb)
        oper.dealiasing(Fb_fft)

        px_rot_fft, py_rot_fft = oper.gradfft_from_fft(rot_fft)
        px_rot = oper.ifft2(px_rot_fft)
        py_rot = oper.ifft2(py_rot_fft)

        px_ux_fft, py_ux_fft = oper.gradfft_from_fft(ux_fft)
        px_ux = oper.ifft2(px_ux_fft)
        py_ux = oper.ifft2(py_ux_fft)

        px_uy_fft, py_uy_fft = oper.gradfft_from_fft(uy_fft)
        px_uy = oper.ifft2(px_uy_fft)
        py_uy = oper.ifft2(py_uy_fft)

        Frot = -ux * px_rot - uy * (py_rot + self.params.beta)
        Frot_fft = oper.fft2(Frot)
        oper.dealiasing(Frot_fft)

        Fx = -ux * px_ux - uy * (py_ux)
        Fx_fft = oper.fft2(Fx)
        oper.dealiasing(Fx_fft)

        Fy = -ux * px_uy - uy * (py_uy)
        Fy_fft = oper.fft2(Fy)
        oper.dealiasing(Fy_fft)

        # Frequency dissipation viscosity
        f_d, f_d_hypo = self.sim.compute_freq_diss()
        freq_diss_EK = f_d + f_d_hypo

        # Energy budget terms. Nonlinear transfer terms, exchange kinetic and
        # potential energy B, dissipation terms.
        transferZ_fft = (
            np.real(rot_fft.conj() * Frot_fft + rot_fft * Frot_fft.conj()) / 2.0
        )
        transferEKu_fft = np.real(ux_fft.conj() * Fx_fft)
        transferEKv_fft = np.real(uy_fft.conj() * Fy_fft)
        B_fft = np.real(uy_fft.conj() * b_fft)

        if self.params.N == 0:
            transferEA_fft = np.zeros_like(transferZ_fft)
        else:
            transferEA_fft = (1 / self.params.N ** 2) * np.real(
                b_fft.conj() * Fb_fft
            )

        dissEKu_fft = np.real(freq_diss_EK * (ux_fft.conj() * ux_fft))
        dissEKv_fft = np.real(freq_diss_EK * (uy_fft.conj() * uy_fft))

        dissEK_fft = np.real(
            freq_diss_EK
            * (
                ux_fft.conj() * ux_fft
                + ux_fft * ux_fft.conj()
                + uy_fft.conj() * uy_fft
                + uy_fft * uy_fft.conj()
            )
            / 2.0
        )
        if self.params.N == 0:
            dissEA_fft = np.zeros_like(dissEK_fft)
        else:
            dissEA_fft = (1 / self.params.N ** 2) * np.real(
                freq_diss_EK * (b_fft.conj() * b_fft)
            )

        transferEK_fft = (
            np.real(
                ux_fft.conj() * Fx_fft
                + ux_fft * Fx_fft.conj()
                + uy_fft.conj() * Fy_fft
                + uy_fft * Fy_fft.conj()
            )
            / 2.0
        )

        # Transfer spectrum 1D Kinetic energy, potential energy and exchange
        # energy
        transferEK_kx, transferEK_ky = self.spectra1D_from_fft(transferEK_fft)
        transferEKu_kx, transferEKu_ky = self.spectra1D_from_fft(transferEKu_fft)
        transferEKv_kx, transferEKv_ky = self.spectra1D_from_fft(transferEKv_fft)
        transferEA_kx, transferEA_ky = self.spectra1D_from_fft(transferEA_fft)
        B_kx, B_ky = self.spectra1D_from_fft(B_fft)

        dissEK_kx, dissEK_ky = self.spectra1D_from_fft(dissEK_fft)
        dissEKu_kx, dissEKu_ky = self.spectra1D_from_fft(dissEKu_fft)
        dissEKv_kx, dissEKv_ky = self.spectra1D_from_fft(dissEKv_fft)
        dissEA_kx, dissEA_ky = self.spectra1D_from_fft(dissEA_fft)

        # Transfer spectrum shell mean
        transferEK_2d = self.spectrum2D_from_fft(transferEK_fft)
        transferEKu_2d = self.spectrum2D_from_fft(transferEKu_fft)
        transferEKv_2d = self.spectrum2D_from_fft(transferEKv_fft)
        transferEA_2d = self.spectrum2D_from_fft(transferEA_fft)
        B_2d = self.spectrum2D_from_fft(B_fft)
        dissEKu_2d = self.spectrum2D_from_fft(dissEKu_fft)
        dissEKv_2d = self.spectrum2D_from_fft(dissEKv_fft)
        dissEA_2d = self.spectrum2D_from_fft(dissEA_fft)
        transferZ_2d = self.spectrum2D_from_fft(transferZ_fft)

        # Dissipation rate at one time
        epsilon_kx = dissEKu_kx.sum() + dissEKv_kx.sum() + dissEA_kx.sum()
        epsilon_ky = dissEKu_ky.sum() + dissEKv_ky.sum() + dissEA_ky.sum()

        # Variables saved in a dictionary
        dict_results = {
            "transferEK_kx": transferEK_kx,
            "transferEK_ky": transferEK_ky,
            "transferEKu_kx": transferEKu_kx,
            "transferEKu_ky": transferEKu_ky,
            "transferEKv_kx": transferEKv_kx,
            "transferEKv_ky": transferEKv_ky,
            "transferEKu_2d": transferEKu_2d,
            "transferEKv_2d": transferEKv_2d,
            "transferEK_2d": transferEK_2d,
            "transferEA_kx": transferEA_kx,
            "transferEA_ky": transferEA_ky,
            "transferEA_2d": transferEA_2d,
            "transferZ_2d": transferZ_2d,
            "B_kx": B_kx,
            "B_ky": B_ky,
            "B_2d": B_2d,
            "dissEK_kx": dissEK_kx,
            "dissEK_ky": dissEK_ky,
            "dissEKu_kx": dissEKu_kx,
            "dissEKu_ky": dissEKu_ky,
            "dissEKu_2d": dissEKu_2d,
            "dissEKv_kx": dissEKv_kx,
            "dissEKv_ky": dissEKv_ky,
            "dissEKv_2d": dissEKv_2d,
            "dissEA_kx": dissEA_kx,
            "dissEA_ky": dissEA_ky,
            "dissEA_2d": dissEA_2d,
            "epsilon_kx": epsilon_kx,
            "epsilon_ky": epsilon_ky,
        }

        if mpi.rank == 0:
            small_value = 3e-10
            for k, v in dict_results.items():
                if k.startswith("transfer"):
                    if abs(v.sum()) > small_value:
                        print("warning: (abs(v.sum()) > small_value) for " + k)
                        print("k = ", k)
                        print("abs(v.sum()) = ", abs(v.sum()))

        return dict_results

    def _online_plot_saving(self, dict_results):

        transfer2D_EA = dict_results["transferEA_2d"]
        transfer2D_EK = dict_results["transferEK_2d"]
        transfer2D_E = transfer2D_EA + transfer2D_EK
        transfer2D_Z = dict_results["transferZ_2d"]
        khE = self.oper.khE
        PiE = cumsum_inv(transfer2D_E) * self.oper.deltak
        PiZ = cumsum_inv(transfer2D_Z) * self.oper.deltak
        self.axe_a.plot(khE + khE[1], PiE, "k")
        self.axe_b.plot(khE + khE[1], PiZ, "g")

    def load_mean(self, tmin=None, tmax=None, keys_to_load=None):
        """Loads data spect_energy_budget.

        It computes the mean between tmin and tmax."""
        means = {}
        with h5py.File(self.path_file, "r") as file:
            times = file["times"][...]
            nt = len(times)
            if tmin is None:
                imin_plot = 0
            else:
                imin_plot = np.argmin(abs(times - tmin))
            if tmax is None:
                imax_plot = nt - 1
            else:
                imax_plot = np.argmin(abs(times - tmax))

            tmin = times[imin_plot]
            tmax = times[imax_plot]

            print(
                "compute mean spectral energy budget\n"
                f"tmin = {tmin:8.6g} ; tmax = {tmax:8.6g}\n"
                f"imin = {imin_plot:8d} ; imax = {imax_plot:8d}"
            )

            for key in list(file.keys()):
                if key.startswith("k"):
                    means[key] = file[key][...]

            if keys_to_load is not None:
                if isinstance(keys_to_load, str):
                    keys_to_load = [keys_to_load]
                for key in keys_to_load:
                    if key not in file.keys():
                        print(key, file.keys())
                        raise ValueError
            else:
                keys_to_load = [
                    key
                    for key in file.keys()
                    if key != "times" and not key.startswith(("k", "info"))
                ]

            for key in keys_to_load:
                spect = file[key][imin_plot : imax_plot + 1].mean(0)
                means[key] = spect
        return means

    def compute_fluxes_mean(self, tmin=None, tmax=None):
        """compute the fluxes mean."""
        # Load data from file
        with h5py.File(self.path_file, "r") as file:
            times = file["times"][...]
            nt = len(times)
            kxE = file["kxE"][...]
            kyE = file["kyE"][...]

            transferEK_kx = file["transferEK_kx"][...]
            transferEK_ky = file["transferEK_ky"][...]
            transferEA_kx = file["transferEA_kx"][...]
            transferEA_ky = file["transferEA_ky"][...]

            dissEK_kx = file["dissEK_kx"][...]
            dissEA_kx = file["dissEA_kx"][...]

            dissEK_ky = file["dissEK_ky"][...]
            dissEA_ky = file["dissEA_ky"][...]

            conv_kx = file["B_kx"][...]
            conv_ky = file["B_ky"][...]

        if tmin is None:
            imin_plot = 0
        else:
            imin_plot = np.argmin(abs(times - tmin))
        if tmax is None:
            imax_plot = nt - 1
        else:
            imax_plot = np.argmin(abs(times - tmax))

        tmin = times[imin_plot]
        tmax = times[imax_plot]

        # compute means kx
        transferEK_kx = transferEK_kx[imin_plot : imax_plot + 1].mean(0)
        transferEA_kx = transferEA_kx[imin_plot : imax_plot + 1].mean(0)

        dissEK_kx = dissEK_kx[imin_plot : imax_plot + 1].mean(0)
        dissEA_kx = dissEA_kx[imin_plot : imax_plot + 1].mean(0)

        id_kx_dealiasing = np.argmin(kxE - self.sim.oper.kxmax_dealiasing) - 1
        id_ky_dealiasing = np.argmin(kyE - self.sim.oper.kymax_dealiasing) - 1

        conv_kx = conv_kx[imin_plot : imax_plot + 1].mean(0)[:id_kx_dealiasing]
        conv_kx = cumsum_inv(conv_kx[1:]) * self.oper.deltakx

        transferEK_kx = transferEK_kx[:id_kx_dealiasing]
        transferEA_kx = transferEA_kx[:id_kx_dealiasing]
        dissEK_kx = dissEK_kx[:id_kx_dealiasing]
        dissEA_kx = dissEA_kx[:id_kx_dealiasing]
        kxE = kxE[:id_kx_dealiasing]

        # Transfer and dissipation kx
        PiEK_kx = cumsum_inv(transferEK_kx[1:]) * self.oper.deltakx
        PiEA_kx = cumsum_inv(transferEA_kx[1:]) * self.oper.deltakx

        DissEK_kx = dissEK_kx[1:].cumsum() * self.oper.deltakx
        DissEA_kx = dissEA_kx[1:].cumsum() * self.oper.deltakx

        # compute means ky
        transferEK_ky = transferEK_ky[imin_plot : imax_plot + 1].mean(0)
        transferEA_ky = transferEA_ky[imin_plot : imax_plot + 1].mean(0)

        dissEK_ky = dissEK_ky[imin_plot : imax_plot + 1].mean(0)
        dissEA_ky = dissEA_ky[imin_plot : imax_plot + 1].mean(0)

        conv_ky = conv_ky[imin_plot : imax_plot + 1].mean(0)[:id_ky_dealiasing]
        conv_ky = cumsum_inv(conv_ky[1:]) * self.oper.deltaky

        transferEK_ky = transferEK_ky[:id_ky_dealiasing]
        transferEA_ky = transferEA_ky[:id_ky_dealiasing]
        dissEK_ky = dissEK_ky[:id_ky_dealiasing]
        dissEA_ky = dissEA_ky[:id_ky_dealiasing]
        kyE = kyE[:id_ky_dealiasing]

        # Transfer and dissipation ky
        PiEK_ky = cumsum_inv(transferEK_ky[1:]) * self.oper.deltaky
        PiEA_ky = cumsum_inv(transferEA_ky[1:]) * self.oper.deltaky

        DissEK_ky = dissEK_ky[1:].cumsum() * self.oper.deltaky
        DissEA_ky = dissEA_ky[1:].cumsum() * self.oper.deltaky

        # save data in dictionary
        fluxes = {
            "kxE": kxE,
            "kyE": kyE,
            "PiEAmean_x": PiEA_kx,
            "PiEKmean_x": PiEK_kx,
            "DissEAmean_x": DissEA_kx,
            "DissEKmean_x": DissEK_kx,
            "PiEAmean_y": PiEA_ky,
            "PiEKmean_y": PiEK_ky,
            "DissEAmean_y": DissEA_ky,
            "DissEKmean_y": DissEK_ky,
            "Conv_x": conv_kx,
            "Conv_y": conv_ky,
        }
        return fluxes

    def plot(
        self, tmin=None, tmax=None, delta_t=2, plot_diss_EK=False, plot_conv=False
    ):
        """Plot the energy budget."""
        # load data from file
        with h5py.File(self.path_file, "r") as file:
            times = file["times"][...]
            nt = len(times)

        if tmin is None:
            imin_plot = 0
        else:
            imin_plot = np.argmin(abs(times - tmin))
        if tmax is None:
            imax_plot = nt - 1
        else:
            imax_plot = np.argmin(abs(times - tmax))

        # Average from tmin and tmax for plot
        delta_t_save = np.mean(times[1:] - times[0:-1])
        delta_i_plot = int(np.round(delta_t / delta_t_save))

        if delta_i_plot == 0 and delta_t != 0.0:
            delta_i_plot = 1
        delta_t = delta_i_plot * delta_t_save

        print(f"plot(tmin={tmin}, tmax={tmax}, delta_t={delta_t:.2f})")

        tmin = times[imin_plot]
        tmax = times[imax_plot]
        print(
            "plot spectral energy budget\n"
            f"tmin = {tmin:8.6g} ; tmax = {tmax:8.6g} ;\n"
            f"delta_t = {delta_t:8.6g} ; imin = {imin_plot:8d} ;\n"
            f"imax = {imax_plot:8d} ; delta_i = {delta_i_plot:8d}"
        )

        fluxes = self.compute_fluxes_mean(tmin, tmax)

        # Parameters of figure 1
        fig, ax1 = self.output.figure_axe()
        ax1.set_xlabel("$k_x/k_{f,x}$")
        ax1.set_ylabel(r"$\Pi(k_x)/\epsilon$")
        ax1.set_xscale("log")
        ax1.set_yscale("linear")
        ax1.set_title("Spectral energy budget\n" + self.output.summary_simul)

        # Parameters of figure 2
        fig, ax2 = self.output.figure_axe()
        ax2.set_xlabel("$k_z/k_{f,z}$")
        ax2.set_ylabel(r"$\Pi(k_z)/\epsilon$")
        ax2.set_xscale("log")
        ax2.set_yscale("linear")
        ax2.set_title("Spectral energy budget\n" + self.output.summary_simul)

        # Compute time average dissipation to normalize fluxes
        spatial_mean_results = self.output.spatial_means.load()

        t = spatial_mean_results["t"]
        epsK_tot = spatial_mean_results["epsK_tot"]
        epsA_tot = spatial_mean_results["epsA_tot"]
        eps_tot = epsK_tot + epsA_tot

        imin_plot_spatial_times = np.argmin(abs(t - tmin))
        imax_plot_spatial_times = np.argmin(abs(t - tmax))

        eps_tot_time_average = eps_tot[
            imin_plot_spatial_times : imax_plot_spatial_times + 1
        ].mean(0)

        # calculate forcing wave-number k_f to normalize k_x and k_z
        pforcing = self.sim.params.forcing
        if pforcing.enable and pforcing.type == "tcrandom_anisotropic":
            nkmax = pforcing.nkmax_forcing
            nkmin = pforcing.nkmin_forcing
            k_f = ((nkmax + nkmin) / 2) * self.sim.oper.deltak
            angle = pforcing.tcrandom_anisotropic.angle
            try:
                if angle.endswith("°"):
                    angle = np.pi / 180 * float(angle[:-1])
            except AttributeError:
                pass
            k_fx = np.sin(angle) * k_f
            k_fy = np.cos(angle) * k_f

            # Band forcing region kx
            k_fxmin = nkmin * self.sim.oper.deltak * np.sin(angle)
            k_fxmin = max(k_fxmin, self.sim.oper.deltakx)
            k_fxmax = nkmax * self.sim.oper.deltak * np.sin(angle)

            # Band forcing region ky
            k_fymin = nkmin * self.sim.oper.deltak * np.cos(angle)
            k_fymin = max(k_fymin, self.sim.oper.deltaky)
            k_fymax = nkmax * self.sim.oper.deltak * np.cos(angle)

            # plot forcing range
            ax1.axvspan(k_fxmin / k_fx, k_fxmax / k_fx, alpha=0.15, color="black")
            ax2.axvspan(k_fymin / k_fy, k_fymax / k_fy, alpha=0.15, color="black")

            # Plot ozmidov scale
            k_o = (self.params.N ** 3 / self.params.forcing.forcing_rate) ** (
                1 / 2
            )
            ax1.axvline(x=k_o / k_fx, color="black", linestyle="--")
            ax2.axvline(x=k_o / k_fy, color="black", linestyle="--")
            ax1.text((k_o / k_fx) + 2, 1, r"$k_o$")
            ax2.text((k_o / k_fy) + 2, 1.7, r"$k_o$")

        else:
            print("Forcing is not enabled k_x and k_z are not normalized")
            k_fx = 1
            k_fy = 1

        # extract variables from the dictionary to plot
        kxE = fluxes["kxE"]
        kyE = fluxes["kyE"]

        PiEK_kx = fluxes["PiEKmean_x"]
        PiEA_kx = fluxes["PiEAmean_x"]

        DissEK_kx = fluxes["DissEKmean_x"]
        DissEA_kx = fluxes["DissEAmean_x"]

        PiEK_ky = fluxes["PiEKmean_y"]
        PiEA_ky = fluxes["PiEAmean_y"]

        DissEK_ky = fluxes["DissEKmean_y"]
        DissEA_ky = fluxes["DissEAmean_y"]

        conv_kx = fluxes["Conv_x"]
        conv_ky = fluxes["Conv_y"]

        def plot_x(x, y, style, label):
            ax1.plot(x[1:] / k_fx, y / eps_tot_time_average, style, label=label)

        plot_x(kxE, (PiEK_kx + PiEA_kx), "k--", r"$\Pi/\epsilon$")
        plot_x(kxE, PiEK_kx, "r", r"$\Pi_K/\epsilon$")
        plot_x(kxE, PiEA_kx, "b", r"$\Pi_A/\epsilon$")
        plot_x(kxE, (DissEK_kx + DissEA_kx), "g", r"$D/\epsilon$")

        if plot_diss_EK:
            plot_x(kxE, DissEK_kx, "r--", r"$D_K$")

        if plot_conv:
            plot_x(kxE, conv_kx, "c", r"$C$")

        ax1.axhline(y=0, color="k", linestyle=":")

        def plot_y(x, y, style, label):
            ax2.plot(x[1:] / k_fy, y / eps_tot_time_average, style, label=label)

        plot_y(kyE, (PiEK_ky + PiEA_ky), "k--", r"$\Pi/\epsilon$")
        plot_y(kyE, PiEK_ky, "r", r"$\Pi_K/\epsilon$")
        plot_y(kyE, PiEA_ky, "b", r"$\Pi_A/\epsilon$")
        plot_y(kyE, (DissEK_ky + DissEA_ky), "g", r"$D/\epsilon$")

        if plot_diss_EK:
            plot_y(kyE, DissEK_ky, "r--", r"$D_K$")

        if plot_conv:
            plot_y(kyE, conv_ky, "c", r"$C$")

        ax2.axhline(y=0, color="k", linestyle=":")

        # Compute k_b: L_b = U / N
        U = np.sqrt(np.mean(abs(self.sim.state.get_var("ux")) ** 2))
        k_b = self.sim.params.N / U
        ax1.axvline(x=k_b / k_fx, color="k", linestyle="--")
        ax2.axvline(x=k_b / k_fy, color="k", linestyle="--")
        ax1.text((k_b / k_fx) + 0.5, 1, r"$k_b$")
        ax2.text((k_b / k_fy) + 0.5, 1.7, r"$k_b$")

        ax1.legend()
        ax2.legend()



"""Mechanism to extend a Simul class with just a simple class
=============================================================

.. autofunction:: extend_simul_class

.. autoclass:: SimulExtender
   :members:
   :private-members:

.. autosummary::
   :toctree:

   spatial_means_regions_milestone

"""

from fluidsim_core.extend_simul import (
    extend_simul_class,
    SimulExtender,
    _extend_simul_class_from_path,
)



import os
import numpy as np
import json
from typing import Dict

import pandas as pd
import xarray as xr
from fluiddyn.util import mpi

from .base import SpecificOutput


def inner_prod(a_fft, b_fft):
    return np.real(a_fft.conj() * b_fft)


class SpatialMeansBase(SpecificOutput):
    """A :class:`SpatialMean` object handles the saving of .

    This class uses the particular functions defined by some solvers
    :func:`` and
    :func``. If the solver doesn't has these
    functions, this class does nothing.
    """

    _tag = "spatial_means"
    _name_file = _tag + ".txt"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "spatial_means"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(tag, attribs={"HAS_TO_PLOT_SAVED": False})

    def __init__(self, output):
        params = output.sim.params

        self.sum_wavenumbers = output.sum_wavenumbers
        try:
            self.vecfft_from_rotfft = output.oper.vecfft_from_rotfft
        except AttributeError:
            pass

        super().__init__(
            output,
            period_save=params.output.periods_save.spatial_means,
            has_to_plot_saved=params.output.spatial_means.HAS_TO_PLOT_SAVED,
        )

        if self.period_save != 0:
            # saved for each initialization to help detecting bugs
            self._save_one_time()
            self.t_last_save = self.sim.time_stepping.t

    def _init_files(self, arrays_1st_time=None):

        if mpi.rank == 0:
            if not os.path.exists(self.path_file):
                self.file = open(self.path_file, "w")
            else:
                self.file = open(self.path_file, "r+")
                # to go to the end of the file
                self.file.seek(0, os.SEEK_END)

    def _online_save(self):
        self()

    def __call__(self):
        """Save the values at one time."""
        tsim = self.sim.time_stepping.t
        if (
            tsim + 1e-15
        ) // self.period_save > self.t_last_save // self.period_save:
            self.t_last_save = self.sim.time_stepping.t
            self._save_one_time()

    def _save_one_time(self, *args):
        self.t_last_save = self.sim.time_stepping.t

    def _init_online_plot(self):
        if mpi.rank == 0:
            width_axe = 0.85
            height_axe = 0.4
            x_left_axe = 0.12
            z_bottom_axe = 0.55

            size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
            fig, axe = self.output.figure_axe(size_axe=size_axe, numfig=3_000_000)
            self.axe_a = axe
            axe.set_xlabel("$t$")
            axe.set_ylabel("$E(t)$")
            axe.set_title("mean quantities\n" + self.output.summary_simul)

            z_bottom_axe = 0.08
            size_axe[1] = z_bottom_axe
            axe = fig.add_axes(size_axe)
            self.axe_b = axe
            axe.set_xlabel("$t$")
            axe.set_ylabel(r"$\epsilon(t)$")

    def load(self):
        dict_results = {}
        return dict_results

    def load_dataset(self, dims=("t",)):
        """Loads results as a xarray dataset."""
        dict_results = self.load()
        # NOTE: format specified in
        # http://xarray.pydata.org/en/stable/generated/xarray.Dataset.from_dict.html
        dset = {"coords": {}, "attrs": {}, "dims": dims, "data_vars": {}}
        for key, value in dict_results.items():
            if isinstance(value, np.ndarray):
                target = "coords" if key in dims else "data_vars"
                dset[target].update({key: {"dims": dims, "data": value}})
            else:
                dset["attrs"].update({key: value})

        return xr.Dataset.from_dict(dset)

    def plot(self):
        pass

    def compute_time_means(self, tstatio=0.0, tmax=None):
        """compute the temporal means."""
        dict_results = self.load()
        times = dict_results["t"]
        imin_mean = np.argmin(abs(times - tstatio))
        imax_mean = None if tmax is None else np.argmin(abs(times - tmax)) + 1

        dict_time_means = {}
        for key, value in dict_results.items():
            if isinstance(value, np.ndarray):
                dict_time_means[key] = np.mean(value[imin_mean:imax_mean])
        return dict_time_means, dict_results

    def _close_file(self):
        try:
            self.file.close()
        except AttributeError:
            pass

    def time_first_saved(self) -> float:
        with open(self.path_file) as file_means:
            line = ""
            while not line.startswith("time ="):
                line = file_means.readline()

        words = line.split()
        return float(words[2])

    def time_last_saved(self) -> float:
        with open(self.path_file, "rb") as file_means:
            nb_char = file_means.seek(0, os.SEEK_END)  # go to the end
            nb_char_to_read = min(nb_char, 1000)
            file_means.seek(-nb_char_to_read, 2)
            line = file_means.readline()
            while line != b"":
                if line.startswith(b"time ="):
                    line_time = line
                line = file_means.readline()

        words = line_time.split()
        return float(words[2])


class SpatialMeansJSON(SpatialMeansBase):
    """Save and load as line-delimited JSON."""

    _tag = "spatial_means"
    _name_file = _tag + ".json"

    def _save_one_time(self, result: Dict[str, float], delimiter: str = "\n"):
        if mpi.rank == 0:
            json.dump(result, self.file)
            self.file.write(delimiter)

        super()._save_one_time()

    def _file_exists(self):
        return self.path_file.endswith(".json") and os.path.exists(self.path_file)

    def load(self):
        if not os.path.exists(self.path_file):
            raise FileNotFoundError(
                f"Spatial means file is missing: {self.path_file}"
            )

        try:
            df = pd.read_json(self.path_file, orient="records", lines=True)
            return df
        except ValueError:
            with open(self.path_file) as fp:
                for line_nb, line in enumerate(fp):
                    line_nb += 1
                    try:
                        json.loads(line)
                    except json.JSONDecodeError:
                        break

            raise IOError(
                f"Error reading spatial means file {self.path_file} in\n\tline"
                f"number {line_nb}:\n\t{line}"
            )

    def load_dataset(self, dims=("t",)):
        df = self.load()
        if isinstance(df, dict):
            return super().load_dataset(dims)
        else:
            df.index = df[dims[0]]

            return xr.Dataset.from_dataframe(df)

    def compute_time_means(self, tstatio=0.0, tmax=None):
        """compute the temporal means."""
        if not self._file_exists():
            return super().compute_time_means(tstatio, tmax)

        df = self.load()
        times = df.t
        imin_mean = abs(times - tstatio).idxmin()
        imax_mean = None if tmax is None else (abs(times - tmax)).idxmin() + 1

        df_mean = df.iloc[imin_mean:imax_mean].mean()
        return df_mean, df

    def time_first_saved(self) -> float:
        if not self._file_exists():
            self.path_file = self.path_file.replace(".json", ".txt")
            return super().time_first_saved()

        with open(self.path_file) as file_means:
            line = file_means.readline()

        result = json.loads(line)
        return result["t"]

    def time_last_saved(self) -> float:
        if not self._file_exists():
            self.path_file = self.path_file.replace(".json", ".txt")
            return super().time_last_saved()

        with open(self.path_file, "rb") as file_means:
            nb_char = file_means.seek(0, os.SEEK_END)  # go to the end
            nb_char_to_read = min(nb_char, 1000)
            file_means.seek(-nb_char_to_read, 2)
            line = file_means.readline()
            while line != b"":
                line_prev = line
                line = file_means.readline()

        result = json.loads(line_prev)
        return result["t"]



import os

import numpy as np
import h5py

from fluiddyn.util import mpi

from .base import SpecificOutput
from .movies import MoviesBase1D


class MoviesSpectra(MoviesBase1D):
    def __init__(self, output, spectra):
        self.spectra = spectra
        super().__init__(output)

    def init_animation(self, *args, **kwargs):
        if "xmax" not in kwargs:
            kwargs["xmax"] = self.oper.khE[-1:][0]
        if "ymax" not in kwargs:
            kwargs["ymax"] = 1.0

        with h5py.File(self.spectra.path_file2D) as file:
            self.times = file["times"][...]

        super().init_animation(*args, **kwargs)

    def get_field_to_plot(self, time, key=None):
        if key is None:
            key = self.key_field
        idx, t_file = self.get_closest_time_file(time)
        with h5py.File(self.spectra.path_file2D) as file:
            y = file["spectrum2D_" + key][idx]
        y[abs(y) < 10e-16] = 0
        return y, t_file

    def get_closest_time_file(self, time):
        """Find the index and value of the closest actual time of the field."""
        idx = np.abs(self.times - time).argmin()
        return idx, self.times[idx]

    def _init_labels(self, xlabel="x"):
        """Initialize the labels."""
        self.ax.set_xlabel(xlabel, fontdict=self.font)
        self.ax.set_ylabel(self.key_field, fontdict=self.font)
        self.ax.set_yscale("log")

    def _get_axis_data(self):
        """Get axis data.

        Returns
        -------

        x : array
          x-axis data.

        """
        with h5py.File(self.spectra.path_file2D) as file:
            x = file["khE"][...]

        return x

    def _set_key_field(self, key_field):
        """
        Defines key_field default.
        """
        if key_field is None:
            key_field = "E"
        self.key_field = key_field


class Spectra(SpecificOutput):
    """Used for the saving of spectra."""

    _tag = "spectra"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "spectra"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(tag, attribs={"HAS_TO_PLOT_SAVED": False})

    def _init_movies(self):
        self.movies = MoviesSpectra(self.output, self)

    def __init__(self, output):
        self.output = output

        if hasattr(self, "_init_movies"):
            self._init_movies()

        params = output.sim.params
        self.nx = int(params.oper.nx)

        self.spectrum2D_from_fft = output.sim.oper.spectrum2D_from_fft
        self.spectra1D_from_fft = output.sim.oper.spectra1D_from_fft

        super().__init__(
            output,
            period_save=params.output.periods_save.spectra,
            has_to_plot_saved=params.output.spectra.HAS_TO_PLOT_SAVED,
        )

    def _init_path_files(self):
        path_run = self.output.path_run
        self.path_file1D = path_run + "/spectra1D.h5"
        self.path_file2D = path_run + "/spectra2D.h5"

    def _init_files(self, arrays_1st_time=None):
        dict_spectra1D, dict_spectra2D = self.compute()
        if mpi.rank == 0:
            if not os.path.exists(self.path_file1D):
                arrays_1st_time = {
                    "kxE": self.sim.oper.kxE,
                    "kyE": self.sim.oper.kyE,
                }
                self._create_file_from_dict_arrays(
                    self.path_file1D, dict_spectra1D, arrays_1st_time
                )
                arrays_1st_time = {"khE": self.sim.oper.khE}
                self._create_file_from_dict_arrays(
                    self.path_file2D, dict_spectra2D, arrays_1st_time
                )
                self.nb_saved_times = 1
            else:
                with h5py.File(self.path_file1D, "r") as file:
                    dset_times = file["times"]
                    self.nb_saved_times = dset_times.shape[0] + 1
                # save the spectra in the file spectra1D.h5
                self._add_dict_arrays_to_file(self.path_file1D, dict_spectra1D)
                # save the spectra in the file spectra2D.h5
                self._add_dict_arrays_to_file(self.path_file2D, dict_spectra2D)

        self.t_last_save = self.sim.time_stepping.t

    def _online_save(self):
        """Save the values at one time. """
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_save >= self.period_save:
            self.t_last_save = tsim
            dict_spectra1D, dict_spectra2D = self.compute()
            if mpi.rank == 0:
                # save the spectra in the file spectra1D.h5
                self._add_dict_arrays_to_file(self.path_file1D, dict_spectra1D)
                # save the spectra in the file spectra2D.h5
                self._add_dict_arrays_to_file(self.path_file2D, dict_spectra2D)
                self.nb_saved_times += 1
                if self.has_to_plot:
                    self._online_plot_saving(dict_spectra1D, dict_spectra2D)

                    if tsim - self.t_last_show >= self.period_show:
                        self.t_last_show = tsim
                        self.axe.get_figure().canvas.draw()

    def compute(self):
        """compute the values at one time."""
        if mpi.rank == 0:
            dict_results = {}
            return dict_results

    def _init_online_plot(self):
        if mpi.rank == 0:
            fig, axe = self.output.figure_axe(numfig=1_000_000)
            self.axe = axe
            axe.set_xlabel("$k_h$")
            axe.set_ylabel("$E(k_h)$")
            axe.set_title("spectra\n" + self.output.summary_simul)

    def _online_plot_saving(self):
        pass

    def load2d_mean(self, tmin=None, tmax=None):
        with h5py.File(self.path_file2D, "r") as file:
            dset_times = file["times"]
            times = dset_times[...]
            nt = len(times)

            kh = file["khE"][...]

            if tmin is None:
                imin_plot = 0
            else:
                imin_plot = np.argmin(abs(times - tmin))

            if tmax is None:
                imax_plot = nt - 1
            else:
                imax_plot = np.argmin(abs(times - tmax))

            tmin = times[imin_plot]
            tmax = times[imax_plot]

            print(
                "compute mean of 2D spectra\n"
                + (
                    "tmin = {0:8.6g} ; tmax = {1:8.6g}"
                    "imin = {2:8d} ; imax = {3:8d}"
                ).format(tmin, tmax, imin_plot, imax_plot)
            )

            dict_results = {"kh": kh}
            for key in list(file.keys()):
                if key.startswith("spectr"):
                    dset_key = file[key]
                    spect = dset_key[imin_plot : imax_plot + 1].mean(0)
                    dict_results[key] = spect
        return dict_results

    def load1d_mean(self, tmin=None, tmax=None):
        with h5py.File(self.path_file1D, "r") as file:
            dset_times = file["times"]
            times = dset_times[...]
            nt = len(times)

            kx = file["kxE"][...]
            # ky = file['kyE'][...]
            kh = kx

            if tmin is None:
                imin_plot = 0
            else:
                imin_plot = np.argmin(abs(times - tmin))

            if tmax is None:
                imax_plot = nt - 1
            else:
                imax_plot = np.argmin(abs(times - tmax))

            tmin = times[imin_plot]
            tmax = times[imax_plot]

            print(
                "compute mean of 1D spectra"
                + (
                    "tmin = {0:8.6g} ; tmax = {1:8.6g}\n"
                    "imin = {2:8d} ; imax = {3:8d}\n"
                ).format(tmin, tmax, imin_plot, imax_plot)
            )

            dict_results = {"kh": kh}
            for key in list(file.keys()):
                if key.startswith("spectr"):
                    dset_key = file[key]
                    spect = dset_key[imin_plot : imax_plot + 1].mean(0)
                    dict_results[key] = spect
        return dict_results

    def plot1d(self):
        raise NotImplementedError

    def plot2d(self):
        raise NotImplementedError



"""Initialization of the field (:mod:`fluidsim.solvers.ad1d.init_fields`)
=========================================================================

.. autoclass:: InitFieldsAD1D
   :members:

.. autoclass:: InitFieldsCos
   :members:

.. autoclass:: InitFieldsGaussian
   :members:

"""
import numpy as np

from fluidsim.base.init_fields import InitFieldsBase, SpecificInitFields


class InitFieldsCos(SpecificInitFields):
    tag = "cos"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)

    # params.init_fields._set_child(cls.tag, attribs={})

    def __call__(self):
        oper = self.sim.oper
        s = np.cos(2 * np.pi * oper.xs / oper.Lx)
        self.sim.state.state_phys[0] = s


class InitFieldsGaussian(SpecificInitFields):
    tag = "gaussian"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)

    # params.init_fields._set_child(cls.tag, attribs={})

    def __call__(self):
        oper = self.sim.oper
        s = np.exp(-((10 * (oper.xs - oper.Lx / 2.0)) ** 2))
        self.sim.state.state_phys[0] = s

        if hasattr(self.sim.state, "statespect_from_statephys"):
            self.sim.state.statespect_from_statephys()


class InitFieldsAD1D(InitFieldsBase):
    """Init the fields for the solver AD1D."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""

        InitFieldsBase._complete_info_solver(
            info_solver, classes=[InitFieldsCos, InitFieldsGaussian]
        )



"""AD1D pseudospectral solver (:mod:`fluidsim.solvers.ad1d.solver`)
===================================================================


Provides:

.. autoclass:: Simul
   :members:
   :private-members:


"""
from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)


class InfoSolverAD1DPseudoSpect(InfoSolverPseudoSpectral):
    def _init_root(self):
        super()._init_root()

        package_base = "fluidsim.solvers.ad1d"
        package = package_base + ".pseudo_spect"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "AD1D"

        classes = self.classes

        classes.Operators.module_name = "fluidsim.operators.operators1d"
        classes.Operators.class_name = "OperatorsPseudoSpectral1D"

        classes.InitFields.module_name = package_base + ".init_fields"
        classes.InitFields.class_name = "InitFieldsAD1D"

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateAD1DPseudoSpectral"

        classes.Output.module_name = "fluidsim.solvers.ad1d.output"
        classes.Output.class_name = "Output"


class Simul(SimulBasePseudoSpectral):
    InfoSolver = InfoSolverAD1DPseudoSpect

    @staticmethod
    def _complete_params_with_default(params):
        SimulBasePseudoSpectral._complete_params_with_default(params)
        params._set_attrib("U", 1.0)

    def tendencies_nonlin(self, state_spect=None, old=None):

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        tendencies_fft[:] = 0.0

        return tendencies_fft

    def compute_freq_complex(self, key):
        assert key == "s_fft"
        omega = 1.0j * self.params.U * self.oper.KX
        return omega


if __name__ == "__main__":
    from scipy.signal import gausspulse

    params = Simul.create_default_params()
    params.output.sub_directory = "examples"

    params.output.periods_save.phys_fields = 1.0
    params.short_name_type_run = "test"
    params.oper.Lx = 10
    params.oper.nx = 256
    params.nu_2 = 1e-3
    params.time_stepping.USE_CFL = False
    params.time_stepping.deltat0 = 1e-3
    params.time_stepping.t_end = 1.0
    params.time_stepping.type_time_scheme = "RK2"
    params.init_fields.type = "in_script"

    sim = Simul(params)

    # Initialize
    x = sim.oper.x
    Lx = sim.oper.Lx
    s = gausspulse(x - Lx / 2, fc=2)
    s_fft = sim.oper.fft(s)

    sim.state.init_statephys_from(s=s)
    sim.state.init_statespect_from(s_fft=s_fft)

    # Plot once
    # sim.output.init_with_initialized_state()
    # sim.output.phys_fields.plot(field='s')

    sim.time_stepping.start()
    # sim.output.phys_fields.plot(field='s')



"""Base module for the output (:mod:`fluidsim.base.output.base`)
================================================================

Provides:

.. autoclass:: OutputBase
   :members:
   :private-members:
   :noindex:

.. autoclass:: OutputBasePseudoSpectral
   :members:
   :private-members:

.. autoclass:: SpecificOutput
   :members:
   :private-members:

.. autoclass:: SimReprMaker
   :members:
   :private-members:
   :noindex:
"""

import datetime
import os
import shutil
import numbers
from copy import copy
from warnings import warn

import numpy as np
import h5py
import matplotlib.pyplot as plt

import fluiddyn
from fluiddyn.util import mpi
from fluiddyn.util import is_run_from_ipython, print_memory_usage
from fluiddyn.io import FLUIDSIM_PATH, Path
from fluidsim_core.output import OutputCore, SimReprMakerCore
from fluidsim_core.params import iter_complete_params

import fluidsim
from fluidsim.util.util import open_patient


class SimReprMaker(SimReprMakerCore):
    """Produce a string representing the simulation"""

    def get_time_as_str(self):
        params = self.sim.params
        if not params.NEW_DIR_RESULTS and (
            params.ONLY_COARSE_OPER or params.init_fields.type == "from_file"
        ):
            return self.time_from_path_run(params.path_run)
        else:
            return super().get_time_as_str()


class OutputBase(OutputCore):
    """Handle the output."""

    SimReprMaker = SimReprMaker

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""
        classes = info_solver.classes.Output._set_child("classes")

        classes._set_child(
            "PrintStdOut",
            attribs={
                "module_name": "fluidsim.base.output.print_stdout",
                "class_name": "PrintStdOutBase",
            },
        )

        classes._set_child(
            "PhysFields",
            attribs={
                "module_name": "fluidsim.base.output.phys_fields2d",
                "class_name": "PhysFieldsBase2D",
            },
        )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        attribs = {
            "ONLINE_PLOT_OK": True,
            "period_refresh_plots": 1,
            "HAS_TO_SAVE": True,
            "sub_directory": "",
        }
        p_output = params._set_child("output", attribs=attribs)

        p_output._set_doc(
            """

See :mod:`fluidsim.output.base`

ONLINE_PLOT_OK: bool (default: True)

    If True, the online plots are enabled.

period_refresh_plots: float (default: 1)

    Period of refreshment of the online plots.

HAS_TO_SAVE: bool (default: True)

    If False, nothing new is saved in the directory of the simulation.

sub_directory: str (default: "")

    A name of a subdirectory where the directory of the simulation is saved.
"""
        )

        p_output._set_child("periods_save")
        p_output.periods_save._set_doc(
            """
Periods (float, in equation time) to set when the specific outputs are saved.
"""
        )
        p_output._set_child("periods_print")
        p_output.periods_print._set_doc(
            """
Periods (float, in equation time) to set when the printing specific outputs are
called.
"""
        )
        p_output._set_child("periods_plot")
        p_output.periods_plot._set_doc(
            """
Periods (float, in equation time) to set when the plots of the specific outputs
are called.
"""
        )

        dict_classes = info_solver.classes.Output.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    def __init__(self, sim):
        super().__init__(sim)

        dict_classes = sim.info.solver.classes.Output.import_classes()

        PrintStdOut = dict_classes["PrintStdOut"]
        self.print_stdout = PrintStdOut(self)

        if not self.params.ONLINE_PLOT_OK:
            for k in self.params.periods_plot._get_key_attribs():
                self.params.periods_plot[k] = 0.0

        if not self._has_to_save:
            for k in self.params.periods_save._get_key_attribs():
                self.params.periods_save[k] = 0.0

    def _init_sim_repr_maker(self):
        sim_repr_maker = super()._init_sim_repr_maker()

        sim = self.sim
        # oper should already be initialized at this point
        if hasattr(self, "oper") and hasattr(self.oper, "_modify_sim_repr_maker"):
            self.oper._modify_sim_repr_maker(sim_repr_maker)

        if hasattr(sim, "_modify_sim_repr_maker"):
            sim._modify_sim_repr_maker(sim_repr_maker)

        # other classes are not initialized at this point
        dict_classes = sim.info_solver.import_classes()
        try:
            cls = dict_classes["Forcing"]
        except KeyError:
            pass
        else:
            if hasattr(cls, "_modify_sim_repr_maker"):
                cls._modify_sim_repr_maker(sim_repr_maker)

        return sim_repr_maker

    def init_with_oper_and_state(self):
        warn(
            "This method has been replaced with a more generic name `post_init`.",
            DeprecationWarning,
        )
        self.post_init()

    def post_init(self):
        sim = self.sim
        super().post_init()

        if mpi.rank == 0:

            objects_to_print = {
                "sim.oper": sim.oper,
                "sim.state": sim.state,
                "sim.time_stepping": sim.time_stepping,
                "sim.init_fields": sim.init_fields,
            }

            if hasattr(sim, "forcing"):
                objects_to_print["sim.forcing"] = sim.forcing

            if hasattr(sim, "preprocess"):
                objects_to_print["sim.preprocess"] = sim.preprocess

            for key, obj in objects_to_print.items():
                self.print_stdout(
                    "{:20s}".format(key + ": ") + str(obj.__class__)
                )

            # print info on the run
            if hasattr(sim.params.time_stepping, "type_time_scheme"):
                specifications = (
                    sim.params.time_stepping.type_time_scheme + " and "
                )
            else:
                specifications = ""
            if mpi.nb_proc == 1:
                specifications += "sequential,\n"
            else:
                specifications += f"parallel ({mpi.nb_proc} proc.)\n"
            self.print_stdout(
                "\nsolver "
                + self.name_solver
                + ", "
                + specifications
                + self.oper.produce_long_str_describing_oper()
                + "path_run =\n"
                + self.path_run
                + "\n"
                + "init_fields.type: "
                + sim.params.init_fields.type
                + "\n"
            )

            if hasattr(self.sim, "produce_str_describing_params"):
                self.print_stdout(
                    "Important parameters: \n"
                    + self.sim.produce_str_describing_params()
                )

        if mpi.rank == 0 and is_run_from_ipython():
            plt.ion()

        if sim.state.is_initialized:
            if hasattr(sim, "forcing") and not sim.forcing.is_initialized():
                return

            self.init_with_initialized_state()

    def _save_info_solver_params_xml(self, replace=False):
        """Save files with information on the solver and on the run."""
        super()._save_info_solver_params_xml(
            replace=replace, comment=f"FluidSim {fluidsim.get_local_version()}"
        )

    def init_with_initialized_state(self):

        if (
            hasattr(self, "_has_been_initialized_with_state")
            and self._has_been_initialized_with_state
        ):
            return

        else:
            self._has_been_initialized_with_state = True

        params = self.sim.params
        # just for the first output
        if (
            hasattr(params.time_stepping, "USE_CFL")
            and params.time_stepping.USE_CFL
        ):
            self.sim.time_stepping.compute_time_increment_CLF()

        if hasattr(self.sim, "forcing") and params.output.HAS_TO_SAVE:
            self.sim.forcing.compute()

        self.print_stdout("Initialization outputs:")

        self.print_stdout.complete_init_with_state()

        dict_classes = copy(self.sim.info.solver.classes.Output.import_classes())

        # The class PrintStdOut has already been instantiated.
        dict_classes.pop("PrintStdOut")

        # to get always the same order (important with mpi)
        keys = sorted(dict_classes.keys())
        classes = [dict_classes[key] for key in keys]

        for Class in classes:
            if mpi.rank == 0:
                self.print_stdout(
                    "{:30s}".format("sim.output." + Class._tag + ":") + str(Class)
                )
            self.__dict__[Class._tag] = Class(self)

        print_memory_usage("\nMemory usage at the end of init. (equiv. seq.)")

        try:
            self.print_size_in_Mo(self.sim.state.state_spect, "state_spect")
        except AttributeError:
            self.print_size_in_Mo(self.sim.state.state_phys, "state_phys")

    def one_time_step(self):

        for k in self.params.periods_print._get_key_attribs():
            period = self.params.periods_print.__dict__[k]
            if period != 0:
                self.__dict__[k]._online_print()

        if self.params.ONLINE_PLOT_OK:
            for k in self.params.periods_plot._get_key_attribs():
                period = self.params.periods_plot.__dict__[k]
                if period != 0:
                    self.__dict__[k]._online_plot()

        if self._has_to_save:
            for k in self.params.periods_save._get_key_attribs():
                period = self.params.periods_save.__dict__[k]
                if period != 0:
                    self.__dict__[k]._online_save()

    def figure_axe(self, numfig=None, size_axe=None):
        if mpi.rank == 0:
            if size_axe is None and numfig is None:
                return plt.subplots()

            if numfig is None:
                fig = plt.figure()
            else:
                fig = plt.figure(numfig)
                fig.clf()
            if size_axe is not None:
                ax = fig.add_axes(size_axe)
            else:
                ax = fig.subplots()
            return fig, ax

    def close_files(self):
        if mpi.rank == 0 and self._has_to_save:
            self.print_stdout.close()
            for k in self.params.periods_save._get_key_attribs():
                period = self.params.periods_save.__dict__[k]
                if period != 0:
                    if hasattr(self.__dict__[k], "_close_file"):
                        self.__dict__[k]._close_file()

    def end_of_simul(self, total_time):
        # self.path_run: str
        path_run = Path(self.path_run)
        self.print_stdout(
            f"Computation completed in {total_time:8.6g} s\n"
            f"path_run =\n{path_run}"
        )
        if self._has_to_save:
            if hasattr(self.sim, "forcing"):
                self.sim.forcing.compute()
            self.one_time_step()
            if self.sim.output.phys_fields.t_last_save < self.sim.time_stepping.t:
                self.phys_fields.save()

        self.close_files()

        if not self.path_run.startswith(FLUIDSIM_PATH):
            path_base = Path(FLUIDSIM_PATH)
            if len(self.params.sub_directory) > 0:
                path_base = path_base / self.params.sub_directory

            new_path_run = path_base / self.sim.name_run

            try:
                if new_path_run.parent.samefile(path_run.parent):
                    return
            except OSError:
                pass

            if mpi.rank == 0 and path_run.exists():
                if not path_base.exists():
                    os.makedirs(path_base)

                shutil.move(self.path_run, path_base)
                print(f"move result directory in directory:\n{new_path_run}")

            self.path_run = str(new_path_run)
            for spec_output in list(self.__dict__.values()):
                if isinstance(spec_output, SpecificOutput):
                    try:
                        spec_output._init_path_files()
                    except AttributeError:
                        pass

            if mpi.nb_proc > 1:
                mpi.comm.barrier()

    def compute_energy(self):
        return 0.0

    def print_size_in_Mo(self, arr, string=None):
        if string is None:
            string = "Size of ndarray (equiv. seq.)"
        else:
            string = "Size of " + string + " (equiv. seq.)"
        mem = arr.nbytes * 1.0e-6
        if mpi.nb_proc > 1:
            mem = mpi.comm.allreduce(mem, op=mpi.MPI.SUM)
        self.print_stdout(string.ljust(30) + f": {mem} Mo")


class OutputBasePseudoSpectral(OutputBase):
    def post_init(self):
        oper = self.oper
        self.sum_wavenumbers = oper.sum_wavenumbers
        super().post_init()

    def compute_energy_fft(self):
        """Compute energy(k)"""
        energy_fft = 0.0
        for k in self.sim.state.keys_state_spect:
            energy_fft += (
                np.abs(self.sim.state.state_spect.get_var(k)) ** 2
            ) / 2.0

        return energy_fft

    def compute_energy(self):
        """Compute the spatially averaged energy."""
        energy_fft = self.compute_energy_fft()
        return self.sum_wavenumbers(energy_fft)


class SpecificOutput:
    """Small class for features useful for specific outputs"""

    def __init__(
        self,
        output,
        period_save=0,
        period_plot=0,
        has_to_plot_saved=False,
        arrays_1st_time=None,
    ):

        sim = output.sim
        params = sim.params

        self.output = output
        self.sim = sim
        self.oper = sim.oper
        self.params = params

        self.period_save = period_save
        self.period_plot = period_plot
        self.has_to_plot = has_to_plot_saved

        if not params.output.ONLINE_PLOT_OK:
            self.period_plot = 0
            self.has_to_plot = False

        if not has_to_plot_saved:
            if self.period_plot > 0:
                self.has_to_plot = True
            else:
                self.has_to_plot = False

        self.period_show = params.output.period_refresh_plots
        self.t_last_show = 0.0

        self._init_path_files()

        if self.has_to_plot:
            self._init_online_plot()

        if not output._has_to_save:
            self.period_save = 0.0

        if self.period_save != 0.0:
            self._init_files(arrays_1st_time)

    def _init_path_files(self):
        if hasattr(self, "_name_file"):
            self.path_file = os.path.join(self.output.path_run, self._name_file)

    def _init_files(self, arrays_1st_time=None):
        if arrays_1st_time is None:
            arrays_1st_time = {}
        dict_results = self.compute()
        if mpi.rank == 0:
            if not os.path.exists(self.path_file):
                self._create_file_from_dict_arrays(
                    self.path_file, dict_results, arrays_1st_time
                )
                self.nb_saved_times = 1
            else:
                with h5py.File(self.path_file, "r") as file:
                    dset_times = file["times"]
                    self.nb_saved_times = dset_times.shape[0] + 1
                self._add_dict_arrays_to_file(self.path_file, dict_results)
        self.t_last_save = self.sim.time_stepping.t

    def _online_save(self):
        """Save the values at one time."""
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_save >= self.period_save:
            self.t_last_save = tsim
            dict_results = self.compute()
            if mpi.rank == 0:
                self._add_dict_arrays_to_file(self.path_file, dict_results)
                self.nb_saved_times += 1
                if self.has_to_plot:
                    self._online_plot_saving(dict_results)
                    if tsim - self.t_last_show >= self.period_show:
                        self.t_last_show = tsim
                        self.fig.canvas.draw()
                        # needed to really show the figures
                        plt.pause(1e-3)

    def _create_file_from_dict_arrays(
        self, path_file, dict_matrix, arrays_1st_time
    ):
        if os.path.exists(path_file):
            print("file NOT created since it already exists!")
        elif mpi.rank == 0:
            with h5py.File(path_file, "w") as file:
                file.attrs["date saving"] = str(datetime.datetime.now()).encode()
                file.attrs["name_solver"] = self.output.name_solver
                file.attrs["name_run"] = self.output.name_run

                self.sim.info._save_as_hdf5(hdf5_parent=file)

                times = np.array([self.sim.time_stepping.t], dtype=np.float64)
                file.create_dataset("times", data=times, maxshape=(None,))

                for k, v in list(arrays_1st_time.items()):
                    file.create_dataset(k, data=v)

                for k, v in list(dict_matrix.items()):
                    if isinstance(v, numbers.Number):
                        arr = np.array([v], dtype=v.__class__)
                        arr.resize((1,))
                        file.create_dataset(k, data=arr, maxshape=(None,))
                    else:
                        arr = np.array(v)
                        arr.resize((1,) + v.shape)
                        file.create_dataset(
                            k, data=arr, maxshape=((None,) + v.shape)
                        )

    def _add_dict_arrays_to_file(self, path_file, dict_matrix):
        if not os.path.exists(path_file):
            raise ValueError("can not add dict arrays in nonexisting file!")

        elif mpi.rank == 0:
            with open_patient(path_file, "r+") as file:
                dset_times = file["times"]
                nb_saved_times = dset_times.shape[0]
                dset_times.resize((nb_saved_times + 1,))
                dset_times[nb_saved_times] = self.sim.time_stepping.t
                for k, v in list(dict_matrix.items()):
                    if isinstance(v, numbers.Number):
                        dset_k = file[k]
                        dset_k.resize((nb_saved_times + 1,))
                        dset_k[nb_saved_times] = v
                    else:
                        dset_k = file[k]
                        dset_k.resize((nb_saved_times + 1,) + v.shape)
                        dset_k[nb_saved_times] = v

    def _add_dict_arrays_to_open_file(self, file, dict_arrays, nb_saved_times):
        if mpi.rank == 0:
            dset_times = file["times"]
            dset_times.resize((nb_saved_times + 1,))
            dset_times[nb_saved_times] = self.sim.time_stepping.t
            for k, v in list(dict_arrays.items()):
                dset_k = file[k]
                dset_k.resize((nb_saved_times + 1, v.size))
                dset_k[nb_saved_times] = v

    def _has_to_online_save(self):
        return (
            self.sim.time_stepping.t + 1e-15
        ) // self.period_save > self.t_last_save // self.period_save

    def _init_online_plot(self):
        pass

    def _online_plot_saving(self, dict_results):
        pass



"""Spectra output (:mod:`fluidsim.solvers.ns2d.output.spectra`)
===============================================================

.. autoclass:: SpectraNS2D
   :members:
   :private-members:

"""

import h5py

import numpy as np

from fluidsim.base.output.spectra import Spectra


class SpectraNS2D(Spectra):
    """Save and plot spectra."""

    def compute(self):
        """compute the values at one time."""
        energy_fft = self.output.compute_energy_fft()
        # compute the spectra 1D
        spectrum1Dkx_E, spectrum1Dky_E = self.spectra1D_from_fft(energy_fft)
        dict_spectra1D = {
            "spectrum1Dkx_E": spectrum1Dkx_E,
            "spectrum1Dky_E": spectrum1Dky_E,
        }
        # compute the spectra 2D
        spectrum2D_E = self.spectrum2D_from_fft(energy_fft)
        dict_spectra2D = {"spectrum2D_E": spectrum2D_E}
        return dict_spectra1D, dict_spectra2D

    def _online_plot_saving(self, dict_spectra1D, dict_spectra2D):
        if (
            self.nx == self.params.oper.ny
            and self.params.oper.Lx == self.params.oper.Ly
        ):
            spectrum2D = dict_spectra2D["spectrum2D_E"]
            khE = self.oper.khE
            coef_norm = khE ** (3.0)
            self.axe.loglog(khE, spectrum2D * coef_norm, "k")
            lin_inf, lin_sup = self.axe.get_ylim()
            if lin_inf < 10e-6:
                lin_inf = 10e-6
            self.axe.set_ylim([lin_inf, lin_sup])
        else:
            print(
                "you need to implement the ploting "
                "of the spectra for this case"
            )

    def plot1d(self, tmin=0, tmax=1000, delta_t=2, coef_compensate=3):

        with h5py.File(self.path_file1D, "r") as h5file:
            dset_times = h5file["times"]

            dset_kxE = h5file["kxE"]
            kh = dset_kxE[...]
            kh2 = kh[:]
            kh2[kh == 0] = 1e-15

            dset_spectrum1Dkx = h5file["spectrum1Dkx_E"]
            dset_spectrum1Dky = h5file["spectrum1Dky_E"]
            times = dset_times[...]

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            delta_t = delta_t_save * delta_i_plot
            if delta_i_plot == 0:
                delta_i_plot = 1

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            print(
                "plot1d(tmin={}, tmax={}, delta_t={:.2f},".format(
                    tmin, tmax, delta_t
                )
                + f" coef_compensate={coef_compensate:.3f})"
            )

            print(
                """plot 1D spectra
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                    tmin_plot,
                    tmax_plot,
                    delta_t,
                    imin_plot,
                    imax_plot,
                    delta_i_plot,
                )
            )

            fig, ax1 = self.output.figure_axe()
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("spectra")
            ax1.set_title("1D spectra\n" + self.output.summary_simul)
            ax1.set_xscale("log")
            ax1.set_yscale("log")

            EKx = dset_spectrum1Dkx[0]
            EKy = dset_spectrum1Dky[0]

            is_asym = len(EKx) == len(EKy)

            coef_norm = kh2 ** (coef_compensate)
            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    EK = dset_spectrum1Dkx[it]
                    if is_asym:
                        EK += dset_spectrum1Dky[it]

                    EK[EK < 10e-16] = 0.0
                    ax1.plot(kh, EK * coef_norm, "k", linewidth=2)

            EK = dset_spectrum1Dkx[imin_plot : imax_plot + 1]
            if is_asym:
                EK += dset_spectrum1Dky[imin_plot : imax_plot + 1]

        EK = EK.mean(0)

        ax1.plot(kh, kh2 ** (-3) * coef_norm, "k", linewidth=1)
        ax1.plot(kh, 0.01 * kh2 ** (-5 / 3) * coef_norm, "k--", linewidth=1)

    def plot2d(self, tmin=0, tmax=1000, delta_t=2, coef_compensate=3):
        with h5py.File(self.path_file2D, "r") as h5file:
            dset_times = h5file["times"]
            # nb_spectra = dset_times.shape[0]
            times = dset_times[...]
            # nt = len(times)

            kh = h5file["khE"][...]
            kh2 = kh[:]
            kh2[kh == 0] = 1e-15

            dset_spectrum = h5file["spectrum2D_E"]

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            print(
                "plot2s(tmin={}, tmax={}, delta_t={:.2f},".format(
                    tmin, tmax, delta_t
                )
                + f" coef_compensate={coef_compensate:.3f})"
            )

            print(
                """plot 2D spectra
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                    tmin_plot,
                    tmax_plot,
                    delta_t,
                    imin_plot,
                    imax_plot,
                    delta_i_plot,
                )
            )

            fig, ax1 = self.output.figure_axe()
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("2D spectra")
            ax1.set_title("2D spectra\n" + self.output.summary_simul)
            ax1.set_xscale("log")
            ax1.set_yscale("log")

            coef_norm = kh2 ** coef_compensate

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    EK = dset_spectrum[it]
                    EK[EK < 10e-16] = 0.0
                    ax1.plot(kh, EK * coef_norm, "k", linewidth=1)

            EK = dset_spectrum[imin_plot : imax_plot + 1].mean(0)
            EK[EK < 10e-16] = 0.0
            ax1.plot(kh, EK * coef_norm, "k", linewidth=2)

            ax1.plot(kh, kh2 ** (-3) * coef_norm, "k--", linewidth=1)
            ax1.plot(kh, 0.01 * kh2 ** (-5.0 / 3) * coef_norm, "k-.", linewidth=1)



"""
Spatial means (:mod:`fluidsim.solvers.plate2d.output.spatial_means`)
==========================================================================


Provides:

.. autoclass:: SpatialMeansPlate2D
   :members:
   :private-members:

"""

import os
import numpy as np


from fluiddyn.util import mpi

from fluidsim.base.output.spatial_means import SpatialMeansBase


class SpatialMeansPlate2D(SpatialMeansBase):
    r"""Compute, save, load and plot spatial means.

    .. |p| mathmacro:: \partial

    If only :math:`W` is forced and dissipated, the energy budget is
    quite simple and can be written as:

    .. math::

       \p_t E_w = - C_{w\rightarrow z} - C_{w\rightarrow \chi} + P_w - D_w,

       \p_t E_z = + C_{w\rightarrow z},

       \p_t E_\chi = + C_{w\rightarrow \chi},

    where

    .. math::

       C_{w\rightarrow z} = \sum_{\mathbf{k}} k^4\mathcal{R}(\hat w \hat z^*),

       C_{w\rightarrow \chi} = -\sum_{\mathbf{k}}
       \mathcal{R}( \widehat{\{ w, z\}} \hat \chi ^* ),

       P_w = \sum_{\mathbf{k}} \mathcal{R}( \hat f_w \hat w^* )

    and

    .. math::

       D_w = 2 \nu_\alpha \sum_{\mathbf{k}} k^{2\alpha} E_w(k).
    """

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim
        (
            Ek_fft,
            El_fft,
            Ee_fft,
            conversion_k_to_l_fft,
            conversion_l_to_e_fft,
        ) = self.output.compute_energies_conversion_fft()

        energy_k = self.sum_wavenumbers(Ek_fft)
        energy_l = self.sum_wavenumbers(El_fft)
        energy_e = self.sum_wavenumbers(Ee_fft)

        conversion_k_to_l = self.sum_wavenumbers(conversion_k_to_l_fft)
        conversion_l_to_e = self.sum_wavenumbers(conversion_l_to_e_fft)

        f_d, f_d_hypo = self.sim.compute_freq_diss()
        epsK = self.sum_wavenumbers(f_d[0] * 2 * Ek_fft)
        epsK_hypo = self.sum_wavenumbers(f_d_hypo[0] * 2 * Ek_fft)

        # assert that z is not dissipated
        assert not (f_d[1].any() and f_d_hypo[1].any())

        if self.sim.params.forcing.enable:
            deltat = self.sim.time_stepping.deltat
            state_spect = self.sim.state.state_spect
            w_fft = state_spect.get_var("w_fft")

            forcing_fft = self.sim.forcing.get_forcing()
            Fw_fft = forcing_fft.get_var("w_fft")

            # assert that z in not forced
            Fz_fft = forcing_fft.get_var("z_fft")
            assert np.allclose(
                abs(Fz_fft).max(), 0.0
            ), "abs(Fz_fft).max(): {}".format(abs(Fz_fft).max())

            P1_fft = np.real(w_fft.conj() * Fw_fft)
            P2_fft = (abs(Fw_fft) ** 2) * deltat / 2
            P1 = self.sum_wavenumbers(P1_fft)
            P2 = self.sum_wavenumbers(P2_fft)

        if mpi.rank == 0:
            energy = energy_k + energy_l + energy_e
            epsK_tot = epsK + epsK_hypo

            self.file.write(f"####\ntime = {tsim:17.13f}\n")
            to_print = (
                "E    = {:31.26e} ; E_k    = {:11.6e} ; "
                "E_l    = {:11.6e} ; E_e    = {:11.6e}\n"
                "epsK = {:11.6e} ; epsK_hypo = {:11.6e} ; "
                "epsK_tot = {:11.6e}\n"
                "conv_k_to_l = {:11.6e} : conv_l_to_e = {:11.6e}\n"
            ).format(
                energy,
                energy_k,
                energy_l,
                energy_e,
                epsK,
                epsK_hypo,
                epsK + epsK_hypo,
                conversion_k_to_l,
                conversion_l_to_e,
            )

            self.file.write(to_print)

            if self.sim.params.forcing.enable:
                to_print = (
                    "P1 = {:11.6e} ; P2 = {:11.6e} ; P_tot = {:11.6e} \n"
                ).format(P1, P2, P1 + P2)
                self.file.write(to_print)

            self.file.flush()
            os.fsync(self.file.fileno())

        if self.has_to_plot and mpi.rank == 0:

            self.axe_a.plot(tsim, energy, "k.")
            self.axe_a.plot(tsim, energy_k, "r.")
            self.axe_a.plot(tsim, energy_l, "b.")
            self.axe_a.plot(tsim, energy_e, "y.")

            self.axe_b.plot(tsim, epsK_tot, "k.")
            self.axe_b.plot(tsim, conversion_k_to_l, "c.")
            self.axe_b.plot(tsim, conversion_l_to_e, "g.")
            if self.sim.params.forcing.enable:
                self.axe_b.plot(tsim, P1 + P2, "m.")

            if tsim - self.t_last_show >= self.period_show:
                self.t_last_show = tsim
                fig = self.axe_a.get_figure()
                fig.canvas.draw()

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}

        with open(self.path_file) as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        lines_P = []
        lines_epsK = []

        for il, line in enumerate(lines):
            if line.startswith("time ="):
                lines_t.append(line)
            if line.startswith("E    ="):
                lines_E.append(line)
            if line.startswith("P1 ="):
                lines_P.append(line)
            if line.startswith("epsK ="):
                lines_epsK.append(line)

        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        t = np.empty(nt)
        E = np.empty(nt)
        E_k = np.empty(nt)
        E_l = np.empty(nt)
        E_e = np.empty(nt)

        if self.sim.params.forcing.enable:
            P1 = np.empty(nt)
            P2 = np.empty(nt)
            P_tot = np.empty(nt)

        epsK = np.empty(nt)
        epsK_hypo = np.empty(nt)
        epsK_tot = np.empty(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            t[il] = float(words[2])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])
            E_k[il] = float(words[6])
            E_l[il] = float(words[10])
            E_e[il] = float(words[14])

            if self.sim.params.forcing.enable:

                line = lines_P[il]
                words = line.split()
                P1[il] = float(words[2])
                P2[il] = float(words[6])
                P_tot[il] = float(words[10])

            line = lines_epsK[il]
            words = line.split()
            epsK[il] = float(words[2])
            epsK_hypo[il] = float(words[6])
            epsK_tot[il] = float(words[10])

        dict_results["t"] = t
        dict_results["E"] = E
        dict_results["E_k"] = E_k
        dict_results["E_l"] = E_l
        dict_results["E_e"] = E_e

        if self.sim.params.forcing.enable:
            dict_results["P1"] = P1
            dict_results["P2"] = P2
            dict_results["P_tot"] = P_tot

        dict_results["epsK"] = epsK
        dict_results["epsK_hypo"] = epsK_hypo
        dict_results["epsK_tot"] = epsK_tot

        return dict_results

    def plot(self, with_dtE=False):
        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]
        E_k = dict_results["E_k"]
        E_l = dict_results["E_l"]
        E_e = dict_results["E_e"]

        epsK = dict_results["epsK"]
        epsK_hypo = dict_results["epsK_hypo"]
        epsK_tot = dict_results["epsK_tot"]

        if with_dtE:
            nt = len(t)
            dtE = np.empty(nt)
            for il in range(nt - 2):
                dtE[il + 1] = (E[il + 2] - E[il]) / (t[il + 2] - t[il])
            dtE[0] = (E[1] - E[0]) / (t[1] - t[0])
            dtE[nt - 1] = (E[nt - 1] - E[nt - 2]) / (t[nt - 1] - t[nt - 2])

        width_axe = 0.85
        height_axe = 0.8
        x_left_axe = 0.12
        z_bottom_axe = 0.15

        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("t")
        ax1.set_ylabel("Energies")
        ax1.plot(t, E, "k", linewidth=2)
        ax1.plot(t, E_k, "r", linewidth=2)
        ax1.plot(t, E_l, "b", linewidth=2)
        ax1.plot(t, E_e, "y", linewidth=2)

        size_axe[1] = z_bottom_axe
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("$t$")
        ax1.set_ylabel(r"$\varepsilon_{tot}$, $P_{tot}$, $\partial_t E$")
        ax1.plot(t, epsK, "k--", linewidth=2)
        ax1.plot(t, epsK_hypo, "k:", linewidth=2)
        ax1.plot(t, epsK_tot, "k", linewidth=2)

        if self.sim.params.forcing.enable:
            P_tot = dict_results["P_tot"]
            ax1.plot(t, P_tot, "m", linewidth=2)

        if with_dtE:
            ax1.plot(t, dtE, "b", linewidth=2)

            should_be_zeros = dtE + epsK_tot
            if self.sim.params.forcing.enable:
                should_be_zeros -= P_tot

            ax1.plot(t, should_be_zeros, "g", linewidth=2)



import unittest

import fluiddyn.util.mpi as mpi

from .predaprey.solver import Simul

from fluidsim.util.testing import TestSimul


@unittest.skipIf(mpi.nb_proc > 1, "plot function works sequentially only")
class TestLorenz(TestSimul):
    Simul = Simul

    @classmethod
    def init_params(cls):
        params = cls.params = cls.Simul.create_default_params()
        params.time_stepping.deltat0 = 0.02
        params.time_stepping.t_end = 0.04
        params.output.periods_print.print_stdout = 0.01

    def test_predaprey(self, params=None):
        """Should be able to run a base experiment."""

        sim = self.sim

        sim.state.state_phys.set_var("X", sim.Xs + 2.0)
        sim.state.state_phys.set_var("Y", sim.Ys + 1.0)

        sim.time_stepping.start()

        sim.output.print_stdout.plot_XY()
        sim.output.print_stdout.plot_XY_vs_time()
        sim.output.print_stdout.plot_deltat()
        sim.output.print_stdout.plot_potential()


if __name__ == "__main__":
    unittest.main()



from fluiddyn.util import mpi

from fluidsim.base.forcing.milestone import (
    ForcingMilestone,
    PeriodicUniform,
    OperatorsPseudoSpectral2D,
)


class ForcingMilestone3D(ForcingMilestone):
    ndim = 3

    def _init_operators(self, sim):
        lx = sim.params.oper.Lx
        ly = sim.params.oper.Ly

        if self.params_milestone.nx_max is None:
            self._is_using_coarse_oper = False
            self.oper_coarse = self.oper2d_big
        else:
            self._is_using_coarse_oper = True

            if mpi.rank == 0:
                params = OperatorsPseudoSpectral2D._create_default_params()
                params.oper.Lx = lx
                params.oper.Ly = ly
                nx = params.oper.nx = self.params_milestone.nx_max
                params.oper.ny = 2 * int(nx * ly / lx / 2)
                params.oper.type_fft = "sequential"
                self.oper_coarse = OperatorsPseudoSpectral2D(params)
                self.oper_coarse_shapeK_loc = self.oper_coarse.shapeK_loc
            else:
                self.oper_coarse = None
                self.oper_coarse_shapeK_loc = None

            if mpi.nb_proc > 1:
                self.oper_coarse_shapeK_loc = mpi.comm.bcast(
                    self.oper_coarse_shapeK_loc, root=0
                )

            self.solid = self.sim.oper.create_arrayX(value=0)
            self.solid_fft = self.sim.oper.create_arrayK(value=0)

    def __init__(self, sim):

        if mpi.rank == 0:
            params = OperatorsPseudoSpectral2D._create_default_params()
            params.oper.Lx = sim.params.oper.Lx
            params.oper.Ly = sim.params.oper.Ly
            params.oper.nx = sim.params.oper.nx
            params.oper.ny = sim.params.oper.ny
            params.oper.type_fft = "sequential"
            self.oper2d_big = OperatorsPseudoSpectral2D(params)
            self.solid2d_big_fft = self.oper2d_big.create_arrayK(value=0)
            self.solid2d_big = self.oper2d_big.create_arrayX(value=0)
        else:
            self.oper2d_big = "sequential"
            self.solid2d_big = None

        super().__init__(sim)

    def _full_from_coarse(self, solid):

        if self._is_using_coarse_oper:
            if mpi.rank == 0:
                solid_coarse_fft = self.oper_coarse.fft(solid)
                nKyc, nKxc = self.oper_coarse_shapeK_loc
                solid_coarse_fft[nKyc // 2, :] = 0.0
                solid_coarse_fft[:, nKxc - 1] = 0.0

                self.oper2d_big.put_coarse_array_in_array_fft(
                    solid_coarse_fft,
                    self.solid2d_big_fft,
                    self.oper_coarse,
                    self.oper_coarse_shapeK_loc,
                )
                self.oper2d_big.ifft_as_arg(
                    self.solid2d_big_fft, self.solid2d_big
                )
            else:
                solid_coarse_fft = None

        else:
            self.solid2d_big = solid

        self.solid = self.sim.oper.build_invariant_arrayX_from_2d_indices12X(
            self.solid2d_big, self.oper2d_big
        )

        return self.solid

    def compute(self, time=None):

        sim = self.sim

        if time is None:
            time = sim.time_stepping.t

        solid, x_coors, y_coors = self.get_solid_field(time)

        solid = self._full_from_coarse(solid)

        vx = sim.state.state_phys.get_var("vx")
        fx = self.sigma * solid * (self.get_speed(time) - vx)
        fx_fft = sim.oper.fft(fx)

        fy = -self.sigma * solid * sim.state.state_phys.get_var("vy")
        fy_fft = sim.oper.fft(fy)

        if sim.params.oper.NO_SHEAR_MODES:
            sim.oper.dealiasing(fx_fft, fy_fft)

        self.fstate.init_statespect_from(vx_fft=fx_fft, vy_fft=fy_fft)


if __name__ == "__main__":

    from time import perf_counter

    import matplotlib.pyplot as plt

    from fluidsim.solvers.ns3d.solver import Simul

    params = Simul.create_default_params()

    diameter = 0.5
    number_cylinders = 3
    speed = 0.1

    ny = params.oper.ny = 96 // 2
    nx = params.oper.nx = 4 * ny / 3

    ly = params.oper.Ly = 3 * diameter * number_cylinders
    lx = params.oper.Lx = ly / ny * nx * 4 / number_cylinders

    params.oper.nz = 8
    params.oper.Lz = 1.0

    params.forcing.enable = True
    params.forcing.type = "milestone"
    params.forcing.milestone.nx_max = 48
    objects = params.forcing.milestone.objects

    objects.number = number_cylinders
    objects.diameter = diameter
    objects.width_boundary_layers = 0.1

    movement = params.forcing.milestone.movement

    # movement.type = "uniform"
    # movement.uniform.speed = 1.0

    # movement.type = "sinusoidal"
    # movement.sinusoidal.length = 14 * diameter
    # movement.sinusoidal.period = 100.0

    movement.type = "periodic_uniform"
    movement.periodic_uniform.length = 14 * diameter
    movement.periodic_uniform.length_acc = 0.25
    movement.periodic_uniform.speed = speed

    params.init_fields.type = "noise"
    params.init_fields.noise.velo_max = 5e-3

    movement = PeriodicUniform(
        speed,
        movement.periodic_uniform.length,
        movement.periodic_uniform.length_acc,
        lx,
    )

    params.time_stepping.t_end = movement.period * 1

    params.nu_8 = 1e-14

    params.output.sub_directory = "milestone"
    # params.output.periods_plot.phys_fields = 5.0
    params.output.periods_print.print_stdout = 10
    params.output.periods_save.phys_fields = 10.0

    sim = Simul(params)

    if mpi.rank == 0 and params.output.periods_plot.phys_fields:
        fig = plt.gcf()
        ax = fig.axes[0]
        ax.axis("equal")
        fig.set_size_inches(14, 4)

    self = milestone = sim.forcing.forcing_maker

    t_start = perf_counter()
    sim.time_stepping.start()
    mpi.printby0(f"Done in {perf_counter() - t_start} s")

    mpi.printby0(
        "For a video, run something like:\n\n"
        f"cd {sim.output.path_run};"
        'ipython -i -c "from fluidsim import load_sim_for_plot as load; '
        "sim=load(); sim.output.phys_fields.animate('vx')\""
    )



"""Time stepping (:mod:`fluidsim.base.dedalus.time_stepping`)
==============================================================

Provides:

.. autoclass:: TimeSteppingDedalus
   :members:
   :private-members:

"""

from dedalus.extras import flow_tools

from fluidsim.base.time_stepping.base import TimeSteppingBase0


class TimeSteppingDedalus(TimeSteppingBase0):
    """Time stepping class to handle Dedalus's event loop and FluidSim output."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        TimeSteppingBase0._complete_params_with_default(params)

    def __init__(self, sim):
        if sim.params.time_stepping.USE_CFL:
            self.init_dedalus_cfl(sim)

        super().__init__(sim)

    def init_dedalus_cfl(self, sim):
        initial_dt = sim.params.time_stepping.deltat0
        if sim.params.ONLY_COARSE_OPER:
            self.CFL = None
            return
        self.CFL = flow_tools.CFL(
            sim.dedalus_solver,
            initial_dt=initial_dt,
            cadence=10,
            safety=1,
            max_change=1.5,
            min_change=0.5,
            max_dt=0.125,
            threshold=0.05,
        )
        self.CFL.add_velocities(("vx", "vz"))

    def start(self):
        self.deltat = self.params.time_stepping.deltat0
        super().start()

    def compute_time_increment_CLF(self):
        if self.CFL is None:
            self.deltat = None
        else:
            self.deltat = self.CFL.compute_dt()

    def one_time_step_computation(self):
        self.deltat = self.sim.dedalus_solver.step(self.deltat)



"""Initialization of the field (:mod:`fluidsim.solvers.ns2d.init_fields`)
=========================================================================

.. autoclass:: InitFieldsNS2D
   :members:

.. autoclass:: InitFieldsNoise
   :members:

.. autoclass:: InitFieldsJet
   :members:

.. autoclass:: InitFieldsDipole
   :members:

"""

import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.init_fields import InitFieldsBase, SpecificInitFields


class InitFieldsNoise(SpecificInitFields):
    """Initialize the state with noise."""

    tag = "noise"

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the `params` container (class method)."""
        super()._complete_params_with_default(params)

        p_noise = params.init_fields._set_child(
            cls.tag, attribs={"velo_max": 1.0, "length": 0.0}
        )

        p_noise._set_doc(
            """
velo_max: float (default 1.)

    Maximum velocity.

length: float (default 0.)

    The smallest (cutoff) scale in the noise.

"""
        )

    def __call__(self):
        rot_fft, ux_fft, uy_fft = self.compute_rotuxuy_fft()
        # energy_fft = 0.5 * (np.abs(ux_fft) ** 2 + np.abs(uy_fft) ** 2)
        self.sim.state.init_from_rotfft(rot_fft)

    def compute_rotuxuy_fft(self):

        params = self.sim.params
        oper = self.sim.oper

        lambda0 = params.init_fields.noise.length
        if lambda0 == 0:
            lambda0 = min(oper.Lx, oper.Ly) / 4.0

        def H_smooth(x, delta):
            return (1.0 + np.tanh(2 * np.pi * x / delta)) / 2.0

        # to compute always the same field... (for 1 resolution...)
        np.random.seed(42 + mpi.rank)

        ux_fft = (
            np.random.random(oper.shapeK)
            + 1j * np.random.random(oper.shapeK)
            - 0.5
            - 0.5j
        )
        uy_fft = (
            np.random.random(oper.shapeK)
            + 1j * np.random.random(oper.shapeK)
            - 0.5
            - 0.5j
        )

        if mpi.rank == 0:
            ux_fft[0, 0] = 0.0
            uy_fft[0, 0] = 0.0

        oper.projection_perp(ux_fft, uy_fft)
        oper.dealiasing(ux_fft, uy_fft)

        k0 = 2 * np.pi / lambda0
        delta_k0 = 1.0 * k0
        ux_fft = ux_fft * H_smooth(k0 - oper.K, delta_k0)
        uy_fft = uy_fft * H_smooth(k0 - oper.K, delta_k0)

        ux = oper.ifft2(ux_fft)
        uy = oper.ifft2(uy_fft)
        velo_max = np.sqrt(ux ** 2 + uy ** 2).max()
        if mpi.nb_proc > 1:
            velo_max = oper.comm.allreduce(velo_max, op=mpi.MPI.MAX)
        ux = params.init_fields.noise.velo_max * ux / velo_max
        uy = params.init_fields.noise.velo_max * uy / velo_max

        ux_fft = oper.fft2(ux)
        uy_fft = oper.fft2(uy)

        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        return rot_fft, ux_fft, uy_fft


class InitFieldsJet(SpecificInitFields):
    """Initialize the state with a jet."""

    tag = "jet"

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the `params` container (class method)."""
        super()._complete_params_with_default(params)

    # params.init_fields._set_child(cls.tag, attribs={})

    def __call__(self):
        oper = self.sim.oper
        rot = self.vorticity_jet()
        rot_fft = oper.fft2(rot)
        rot_fft[oper.K == 0] = 0.0
        self.sim.state.init_from_rotfft(rot_fft)

    def vorticity_jet(self):
        oper = self.sim.oper
        Ly = oper.Ly
        a = 0.5
        b = Ly / 2.0
        omega0 = 2.0
        omega = omega0 * (
            np.exp(-(((oper.YY - Ly / 2.0 + b / 2.0) / a) ** 2))
            - np.exp(-(((oper.YY - Ly / 2.0 - b / 2.0) / a) ** 2))
            + np.exp(-(((oper.YY - Ly / 2.0 + b / 2.0 + Ly) / a) ** 2))
            - np.exp(-(((oper.YY - Ly / 2.0 - b / 2.0 + Ly) / a) ** 2))
            + np.exp(-(((oper.YY - Ly / 2.0 + b / 2.0 - Ly) / a) ** 2))
            - np.exp(-(((oper.YY - Ly / 2.0 - b / 2.0 - Ly) / a) ** 2))
        )
        return omega


class InitFieldsDipole(SpecificInitFields):
    """Initialize the state with a dipole."""

    tag = "dipole"

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the `params` container (class method)."""
        super()._complete_params_with_default(params)

    # params.init_fields._set_child(cls.tag, attribs={})

    def __call__(self):
        rot = self.vorticity_shape_1dipole()
        rot_fft = self.sim.oper.fft2(rot)
        self.sim.state.init_from_rotfft(rot_fft)

    def vorticity_shape_1dipole(self):
        oper = self.sim.oper
        xs = oper.Lx / 2.0
        ys = oper.Ly / 2.0
        theta = np.pi / 2.3
        b = 2.5
        omega = np.zeros(oper.shapeX_loc)

        def wz_2LO(XX, YY, b):
            return 2 * np.exp(-(XX ** 2 + (YY - b / 2.0) ** 2)) - 2 * np.exp(
                -(XX ** 2 + (YY + b / 2.0) ** 2)
            )

        for ip in range(-1, 2):
            for jp in range(-1, 2):
                XX_s = np.cos(theta) * (oper.XX - xs - ip * oper.Lx) + np.sin(
                    theta
                ) * (oper.YY - ys - jp * oper.Ly)
                YY_s = np.cos(theta) * (oper.YY - ys - jp * oper.Ly) - np.sin(
                    theta
                ) * (oper.XX - xs - ip * oper.Lx)
                omega = omega + wz_2LO(XX_s, YY_s, b)
        return omega


class InitFieldsNS2D(InitFieldsBase):
    """Initialize the state for the solver NS2D."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""
        InitFieldsBase._complete_info_solver(
            info_solver,
            classes=[InitFieldsNoise, InitFieldsJet, InitFieldsDipole],
        )



"""Forcing schemes (:mod:`fluidsim.base.forcing.base`)
============================================================


Provides:

.. autoclass:: ForcingBase
   :members:
   :private-members:

.. autoclass:: ForcingBasePseudoSpectral
   :members:
   :private-members:

"""

from functools import partial

import numpy as np

from fluiddyn.util import mpi
from fluidsim_core.params import iter_complete_params

from .specific import (
    InScriptForcingPseudoSpectral,
    InScriptForcingPseudoSpectralCoarse,
    Proportional,
    TimeCorrelatedRandomPseudoSpectral,
)


class ForcingBase:
    """Organize the forcing schemes (base class)"""

    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        """Complete the ParamContainer info_solver."""
        classesXML = info_solver.classes.Forcing._set_child("classes")

        if classes is not None:
            for cls in classes:
                classesXML._set_child(
                    cls.tag,
                    attribs={
                        "module_name": cls.__module__,
                        "class_name": cls.__name__,
                    },
                )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        params._set_child(
            "forcing",
            attribs={
                "enable": False,
                "type": "",
                "available_types": [],
                "forcing_rate": 1.0,
                "key_forced": None,
            },
        )

        params.forcing._set_doc(
            """

See :mod:`fluidsim.base.forcing`.

type : str

  Type of forcing.

available_types : list

  Available types that can be used.

forcing_rate : float

  Forcing injection rate.

key_forced: {None} or str

  The key of the variable to be forced. If it is None, a default key depending
  of the type of forcing is used.

"""
        )

        dict_classes = info_solver.classes.Forcing.import_classes()
        # iter on the dict in a determined order
        iter_complete_params(
            params,
            info_solver,
            (dict_classes[key] for key in sorted(dict_classes)),
        )

    @staticmethod
    def _modify_sim_repr_maker(sim_repr_maker):
        sim = sim_repr_maker.sim
        if not sim.params.forcing.enable:
            return
        dict_classes = sim.info_solver.classes.Forcing.import_classes()
        try:
            cls = dict_classes[sim.params.forcing.type]
        except KeyError:
            return
        if hasattr(cls, "_modify_sim_repr_maker"):
            return cls._modify_sim_repr_maker(sim_repr_maker)

    def __init__(self, sim):
        params = sim.params
        self.type_forcing = params.forcing.type

        dict_classes = sim.info.solver.classes.Forcing.import_classes()

        if self.type_forcing not in dict_classes:

            # temporary trick to open old simulations
            if self.type_forcing == "random" and "tcrandom" in dict_classes:
                self.type_forcing = "tcrandom"
                params.forcing.__dict__["tcrandom"] = params.forcing["random"]
            else:
                if mpi.rank == 0:
                    print("dict_classes:", dict_classes)
                raise ValueError(
                    "Wrong value for params.forcing.type: " + self.type_forcing
                )

        ClassForcing = dict_classes[self.type_forcing]

        self.sim = sim
        if not params.ONLY_COARSE_OPER:
            self.forcing_maker = ClassForcing(sim)
        else:
            self.forcing_maker = None

        if hasattr(ClassForcing, "get_info"):
            self.get_info = partial(ClassForcing.get_info, sim)

        self._t_last_computed = -np.inf

    def __call__(self, key):
        """Return the variable corresponding to the given key."""
        forcing = self.get_forcing()
        return forcing.get_var(key)

    def compute(self):
        time = self.sim.time_stepping.t
        if time > self._t_last_computed:
            self.forcing_maker.compute()
            self._t_last_computed = time

    def get_forcing(self):
        return self.forcing_maker.forcing_phys

    def is_initialized(self):
        if hasattr(self.forcing_maker, "is_initialized"):
            return self.forcing_maker.is_initialized

        else:
            return True


class ForcingBasePseudoSpectral(ForcingBase):
    """Organize the forcing schemes (pseudo-spectra)

    .. inheritance-diagram:: ForcingBasePseudoSpectral

    """

    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        """Complete the ParamContainer info_solver."""

        classes_default = (
            InScriptForcingPseudoSpectral,
            InScriptForcingPseudoSpectralCoarse,
            Proportional,
            TimeCorrelatedRandomPseudoSpectral,
        )

        if classes is None:
            classes = classes_default[:]
        else:
            for cls_default in classes_default:
                if not any(cls_default.tag == cls.tag for cls in classes):
                    classes.append(cls_default)

        ForcingBase._complete_info_solver(info_solver, classes=classes)

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        ForcingBase._complete_params_with_default(params, info_solver)

        params.forcing._set_attribs({"nkmax_forcing": 5, "nkmin_forcing": 4})

    def get_forcing(self):
        return self.forcing_maker.forcing_fft



"""Operators for Basilisk (:mod:`fluidsim.base.basilisk.operators`)
===================================================================

Provides:

.. autoclass:: OperatorsBasilisk2D
   :members:
   :private-members:

"""

import numpy as np

import basilisk.stream as basilisk


class OperatorsBasilisk2D:
    """2D operators."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""

        attribs = {"nx": 48, "ny": 48, "Lx": 8, "Ly": 8}
        params._set_child("oper", attribs=attribs)

    def __init__(self, params=None, SEQUENTIAL=None):

        self.basilisk = basilisk

        nx = self.nx_seq = self.ny_seq = int(params.oper.nx)

        self.basilisk.init_grid(nx)

        self.params = params
        self.shapeX_seq = self.shapeX_loc = [nx, nx]

        self.x_seq = x = np.linspace(0, 1, nx)
        self.y_seq = y = np.linspace(0, 1, nx)
        self.X, self.Y = np.meshgrid(x, y)

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""

        return ""

    def produce_long_str_describing_oper(self):
        """Produce a string describing the operator."""

        return "2d Basilisk simulation\n"
